"""Initializing a Kubernetes client."""

from __future__ import annotations

import os

from kubernetes_asyncio import config

__all__ = ["initialize_kubernetes"]


async def initialize_kubernetes() -> None:
    """Load the Kubernetes configuration.

    This has to be run once per process and should be run during application
    startup.  This function handles Kubernetes configuration independent of
    any given Kubernetes client so that clients can be created for each
    request.

    Notes
    -----
    If ``KUBERNETES_PORT`` is set in the environment, this will use
    ``load_incluster_config`` to get configuration information from the local
    pod metadata.  Otherwise, it will use ``load_kube_config`` to read
    configuration from the user's home directory.
    """
    if "KUBERNETES_PORT" in os.environ:
        config.load_incluster_config()
    else:
        await config.load_kube_config()



"""Tests for FastAPI helper code."""

from __future__ import annotations

import pytest
from fastapi import FastAPI, status
from httpx import ASGITransport, AsyncClient

from safir.fastapi import ClientRequestError, client_request_error_handler
from safir.models import ErrorLocation, ErrorModel


@pytest.mark.asyncio
async def test_client_request_error() -> None:
    app = FastAPI()
    app.exception_handler(ClientRequestError)(client_request_error_handler)

    class UnknownUserError(ClientRequestError):
        error = "unknown_user"
        status_code = status.HTTP_404_NOT_FOUND

    class InvalidAddressError(ClientRequestError):
        error = "invalid_address"

    class PermissionDeniedError(ClientRequestError):
        error = "permission_denied"
        status_code = status.HTTP_403_FORBIDDEN

    @app.get("/user")
    async def user_error() -> dict[str, str]:
        raise UnknownUserError("Unknown user", ErrorLocation.path, ["user"])

    @app.get("/address")
    async def address_error() -> dict[str, str]:
        try:
            raise InvalidAddressError("Invalid address")
        except ClientRequestError as e:
            e.location = ErrorLocation.body
            e.field_path = ["user", "address"]
            raise

    @app.get("/permission")
    async def permission_error() -> dict[str, str]:
        raise PermissionDeniedError("Permission denied")

    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/user")
        assert r.status_code == 404
        assert r.json() == {
            "detail": [
                {
                    "msg": "Unknown user",
                    "type": "unknown_user",
                    "loc": ["path", "user"],
                }
            ]
        }

        r = await client.get("/address")
        assert r.status_code == 422
        error = ErrorModel.model_validate(r.json())
        assert error.detail[0].loc == [ErrorLocation.body, "user", "address"]
        assert error.detail[0].msg == "Invalid address"
        assert error.detail[0].type == "invalid_address"

        r = await client.get("/permission")
        assert r.status_code == 403
        assert r.json() == {
            "detail": [
                {"msg": "Permission denied", "type": "permission_denied"}
            ]
        }



"""Helpers for app metrics testing."""

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import timedelta

from aiokafka.admin.client import NewTopic

from safir.metrics import EventsConfiguration, KafkaMetricsConfiguration
from safir.metrics._event_manager import KafkaEventManager

from .kafka import KafkaClients, KafkaStack

__all__ = ["MetricsStack", "make_metrics_stack"]


@dataclass
class MetricsStack:
    """Objects and external services in a full app metrics stack."""

    kafka_stack: KafkaStack
    """Kafka containers and connection info."""

    kafka_clients: KafkaClients
    """Kafka clients that point at the Kafka stack."""

    kafka_max_batch_size: int
    """The maximum number of messages to allow in an aiokafka produce batch.

    If another message is published while there are this many messages in the
    batch, aoikafka will block to send the batch.
    """

    event_manager: KafkaEventManager
    """An uninialized Kafka event manager."""

    metrics_config: KafkaMetricsConfiguration
    """Metrics configuration for all of the services in the stack."""


@asynccontextmanager
async def make_metrics_stack(
    *,
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
    kafka_max_batch_size: int = 16384,
) -> AsyncGenerator[MetricsStack]:
    """Yield a full metrics stack with underlying infrastructure.

    Parameters
    ----------
    kafka_stack
        A bucket of containers.
    kafka_clients
        A bucket of clients that point at those containers.
    kafka_max_batch_size
        The maximum number of messages to allow in an aiokafka produce batch.
        If another message is published while there are this many messages in
        the batch, aoikafka will block to send the batch.
    """
    config = KafkaMetricsConfiguration(
        application="testapp",
        enabled=True,
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_stack.kafka_connection_settings,
        schema_manager=kafka_stack.schema_manager_settings,
        # This should be longer than it takes to restart the Kafka
        # container in a test. We're going to fake time in the test so we
        # can advance it quickly
        backoff_interval=timedelta(hours=1),
        raise_on_error=False,
    )

    manager = config.make_manager(
        extra_broker_settings={
            "max_batch_size": kafka_max_batch_size,
        }
    )
    try:
        topic = NewTopic(
            name="what.ever.testapp",
            num_partitions=1,
            replication_factor=1,
        )
        await kafka_clients.kafka_admin_client.create_topics([topic])

        await manager.initialize()

        yield MetricsStack(
            kafka_stack=kafka_stack,
            kafka_clients=kafka_clients,
            kafka_max_batch_size=kafka_max_batch_size,
            event_manager=manager,
            metrics_config=config,
        )
    finally:
        await manager.aclose()



"""Test `safir.dependencies.arq`."""

from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Annotated, Any

import pytest
from arq.constants import default_queue_name
from asgi_lifespan import LifespanManager
from fastapi import Depends, FastAPI, HTTPException
from httpx import ASGITransport, AsyncClient

from safir.arq import ArqMode, JobNotFound, JobResultUnavailable, MockArqQueue
from safir.dependencies.arq import arq_dependency


@asynccontextmanager
async def _lifespan(app: FastAPI) -> AsyncGenerator[None]:
    await arq_dependency.initialize(mode=ArqMode.test, redis_settings=None)
    yield


@pytest.mark.asyncio
async def test_arq_dependency_mock() -> None:
    """Test the arq dependency entirely through the MockArqQueue."""
    app = FastAPI(lifespan=_lifespan)

    @app.post("/")
    async def post_job(
        *,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> dict[str, Any]:
        """Create a job."""
        job = await arq_queue.enqueue("test_task", "hello", a_number=42)
        return {
            "job_id": job.id,
            "job_status": job.status,
            "job_name": job.name,
            "job_args": job.args,
            "job_kwargs": job.kwargs,
            "job_queue_name": job.queue_name,
        }

    @app.get("/jobs/{job_id}")
    async def get_metadata(
        *,
        job_id: str,
        queue_name: str | None = None,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> dict[str, Any]:
        """Get metadata about a job."""
        try:
            job = await arq_queue.get_job_metadata(
                job_id, queue_name=queue_name
            )
        except JobNotFound as e:
            raise HTTPException(status_code=404) from e
        return {
            "job_id": job.id,
            "job_status": job.status,
            "job_name": job.name,
            "job_args": job.args,
            "job_kwargs": job.kwargs,
            "job_queue_name": job.queue_name,
        }

    @app.get("/results/{job_id}")
    async def get_result(
        *,
        job_id: str,
        queue_name: str | None = None,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> dict[str, Any]:
        """Get the results for a job."""
        try:
            job_result = await arq_queue.get_job_result(
                job_id, queue_name=queue_name
            )
        except (JobNotFound, JobResultUnavailable) as e:
            raise HTTPException(status_code=404, detail=str(e)) from e

        # For testing purposes, turn exceptions into something serializable.
        if isinstance(job_result.result, BaseException):
            job_result.result = f"EXCEPTION {type(job_result.result).__name__}"
        return {
            "job_id": job_result.id,
            "job_status": job_result.status,
            "job_name": job_result.name,
            "job_args": job_result.args,
            "job_kwargs": job_result.kwargs,
            "job_queue_name": job_result.queue_name,
            "job_result": job_result.result,
            "job_success": job_result.success,
        }

    @app.post("/jobs/{job_id}/inprogress")
    async def post_job_inprogress(
        *,
        job_id: str,
        queue_name: str | None = None,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> None:
        """Toggle a job to in-progress, for testing."""
        try:
            await arq_queue.set_in_progress(job_id, queue_name=queue_name)
        except JobNotFound as e:
            raise HTTPException(status_code=404, detail=str(e)) from e

    @app.post("/jobs/{job_id}/complete")
    async def post_job_complete(
        *,
        job_id: str,
        queue_name: str | None = None,
        result: str | None = None,
        success: bool = True,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> None:
        """Toggle a job to complete, for testing."""
        try:
            await arq_queue.set_complete(
                job_id, result=result, success=success, queue_name=queue_name
            )
        except JobNotFound as e:
            raise HTTPException(status_code=404, detail=str(e)) from e

    @app.post("/jobs/{job_id}/abort")
    async def abort_job(
        *,
        job_id: str,
        queue_name: str | None = None,
        arq_queue: Annotated[MockArqQueue, Depends(arq_dependency)],
    ) -> None:
        """Abort a job, for testing."""
        try:
            await arq_queue.abort_job(job_id, queue_name=queue_name)
        except JobNotFound as e:
            raise HTTPException(status_code=404, detail=str(e)) from e

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with LifespanManager(app):
        async with AsyncClient(transport=transport, base_url=base_url) as c:
            r = await c.post("/")
            assert r.status_code == 200
            data = r.json()
            job_id = data["job_id"]
            assert data["job_status"] == "queued"
            assert data["job_name"] == "test_task"
            assert data["job_args"] == ["hello"]
            assert data["job_kwargs"] == {"a_number": 42}
            assert data["job_queue_name"] == default_queue_name

            r = await c.get(f"/jobs/{job_id}")
            assert r.status_code == 200
            assert data["job_kwargs"] == {"a_number": 42}

            # Wrong queue name
            r = await c.get(f"/jobs/{job_id}?queue_name=queue2")
            assert r.status_code == 404

            # Result should not be available
            r = await c.get(f"/results/{job_id}")
            assert r.status_code == 404
            data = r.json()
            assert data["detail"] == (
                f"Job result could not be found. id={job_id}"
            )

            # Set to in-progress
            r = await c.post(f"/jobs/{job_id}/inprogress")
            r = await c.get(f"/jobs/{job_id}")
            data = r.json()
            assert data["job_status"] == "in_progress"

            # Set to successful completion
            r = await c.post(f"/jobs/{job_id}/complete?result=done")
            r = await c.get(f"/results/{job_id}")
            assert r.status_code == 200
            data = r.json()
            assert data["job_status"] == "complete"
            assert data["job_result"] == "done"
            assert data["job_success"] is True

            # Aborting a completed job does nothing.
            r = await c.post(f"/jobs/{job_id}/abort")
            r = await c.get(f"/results/{job_id}")
            assert r.status_code == 200
            data = r.json()
            assert data["job_status"] == "complete"

            # Create a new job and abort it before starting it, which should
            # delete the job.
            r = await c.post("/")
            assert r.status_code == 200
            data = r.json()
            job_id = data["job_id"]
            r = await c.get(f"/jobs/{job_id}")
            assert r.status_code == 200
            r = await c.post(f"/jobs/{job_id}/abort")
            r = await c.get(f"/results/{job_id}")
            assert r.status_code == 404

            # Create a new job, start it, and then abort it. This should keep
            # the job but mark it complete and failed.
            r = await c.post("/")
            assert r.status_code == 200
            data = r.json()
            job_id = data["job_id"]
            r = await c.get(f"/jobs/{job_id}")
            assert r.status_code == 200
            r = await c.post(f"/jobs/{job_id}/inprogress")
            r = await c.post(f"/jobs/{job_id}/abort")
            r = await c.get(f"/results/{job_id}")
            assert r.status_code == 200
            data = r.json()
            assert data["job_status"] == "complete"
            assert data["job_result"] == "EXCEPTION CancelledError"
            assert data["job_success"] is False



"""Docker container for testing with a Kafka Confluent schema registry."""

from __future__ import annotations

from typing import Any, Self

import httpx
from httpx import ReadError, RemoteProtocolError
from testcontainers.core.container import DockerContainer
from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import wait_container_is_ready

from ._constants import CONFLUENT_VERSION_TAG, SCHEMA_REGISTRY_REPOSITORY

__all__ = ["SchemaRegistryContainer"]


class SchemaRegistryContainer(DockerContainer):
    """A Testcontainers_ Confluent schema registry container.

    Parameters
    ----------
    network
        Docker network to put this container in. The Kafa container that will
        serve as storage must be on the same network.
    kafka_bootstrap_servers
        Comma-separated list of Kafka boostrap servers for the Kafka instance
        that will serve as storage.
    image
        Docker image to use.
    """

    def __init__(
        self,
        network: Network,
        kafka_bootstrap_servers: str = "kafka:9092",
        image: str = f"{SCHEMA_REGISTRY_REPOSITORY}:{CONFLUENT_VERSION_TAG}",
        **kwargs: dict[str, Any],
    ) -> None:
        super().__init__(image, **kwargs)
        self.port = 8081
        self.with_exposed_ports(self.port)
        self.with_env(
            "SCHEMA_REGISTRY_KAFKASTORE_BOOTSTRAP_SERVERS",
            kafka_bootstrap_servers,
        )
        self.with_env("SCHEMA_REGISTRY_LISTENERS", "http://0.0.0.0:8081")
        self.with_env(
            "SCHEMA_REGISTRY_HOST_NAME", self.get_docker_client().host()
        )

    def start(self, *args: list[Any], **kwargs: dict[str, Any]) -> Self:
        """Start the container.

        Wait for health after starting the container. Any arguments are
        passed verbatim to the standard testcontainers start method.
        """
        super().start(*args, **kwargs)
        self.health()
        return self

    def get_url(self) -> str:
        """Construct the URL to the schema registry."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        return f"http://{host}:{port}"

    def reset(self) -> None:
        """Reset the contents of the schema registry to its initial state."""
        url = f"{self.get_url()}/subjects"
        subjects = httpx.get(url).json()
        for subject in subjects:
            httpx.delete(f"{url}/{subject}")
            httpx.delete(f"{url}/{subject}?permanent=true")

    @wait_container_is_ready(ReadError, RemoteProtocolError)
    def health(self) -> None:
        """Check the health of the container.

        Consider the container healthy once the schema registry can be
        contacted from the host network.
        """
        url = f"{self.get_url()}/subjects"
        httpx.get(url, timeout=5).raise_for_status()



"""Tools for publishing events from appliactions for later analysis."""

import functools
import time
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta
from enum import StrEnum, auto
from typing import Concatenate, cast, override
from uuid import uuid4

import structlog
from aiokafka.admin.client import AIOKafkaAdminClient
from dataclasses_avroschema.pydantic import AvroBaseModel
from faststream.kafka import KafkaBroker
from faststream.kafka.publisher.asyncapi import AsyncAPIDefaultPublisher
from pydantic import create_model
from structlog.stdlib import BoundLogger

from ..kafka import Compatibility, PydanticSchemaManager, SchemaInfo
from ._constants import EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL
from ._exceptions import (
    DuplicateEventError,
    EventManagerUnintializedError,
    EventManagerUsageError,
    KafkaTopicError,
)
from ._models import EventMetadata, EventPayload
from ._testing import PublishedList

__all__ = [
    "EventManager",
    "EventPublisher",
    "FailedEventPublisher",
    "KafkaEventManager",
    "KafkaEventPublisher",
    "MockEventManager",
    "MockEventPublisher",
    "NoopEventManager",
    "NoopEventPublisher",
]


class _State(StrEnum):
    error = auto()
    initialized = auto()
    uninitialized = auto()


@dataclass
class _CallTracker:
    "Track the status and time that an operation was attempted."

    last_state = _State.uninitialized
    failed_at = datetime(1900, 1, 1, tzinfo=UTC)


def abandonable[**P, R, T: "KafkaEventManager"](
    func: Callable[Concatenate[T, P], Coroutine[None, None, R]],
) -> Callable[Concatenate[T, P], Coroutine[None, None, R | None]]:
    """Catch exceptions from wrapped function and log errors instead.

    Doesn't catch exceptions that are subclasses of EventManagerUsageError.
    Changes the original return value R into the union R | None.

    This function is meant to be used as a decorator on KafkaEventManager
    methods. If an exception is caught, puts the KafkaEventManager instance
    into an error state, logs an error, and returns None. In this error state,
    subsequent calls to this method will just return None instantly. The error
    state will be reset after a configurable amount of time and the call will
    be tried again.
    """
    call_tracker = _CallTracker()

    async def wrapper(self: T, *args: P.args, **kwargs: P.kwargs) -> R | None:
        logger = self.logger.bind(attempted_operation=func.__name__)
        now = datetime.now(tz=UTC)

        if self._state == _State.error:
            elapsed = now - call_tracker.failed_at

            # If we still have backoff time remaining, just return None.
            if elapsed <= self._backoff_interval:
                return None

            # If we're past our backoff time, reset our error state and try the
            # call. This might put us in an error state again.
            self._state = call_tracker.last_state
            logger.info("Trying app metrics operation again")

        try:
            # Try to call the original method
            return await func(self, *args, **kwargs)
        except EventManagerUsageError:
            raise
        except Exception:
            if self._raise_on_error:
                raise

            call_tracker.failed_at = now
            call_tracker.last_state = self._state
            self._state = _State.error

            logger.exception(
                f"App metrics collection operation failed. This may be due"
                f" to instability in the underlying metrics infrastructure"
                f" like Kafka. Any further attempts at app metrics operations"
                f" will be no-ops until"
                f" {self._backoff_interval.total_seconds()} seconds from now.",
                backoff_seconds_remaining=self._backoff_interval.total_seconds(),
            )
        return None

    # functools.wraps injects Any's into the wrapped function definition for
    # reasons I don't quite understand, and it changes the return type to
    # _Wrapped:
    # https://github.com/python/mypy/issues/17171
    # https://github.com/python/mypy/issues/18204
    #
    # The fix here is to mutate the wrapped function just like functools.wraps
    # does, but to mutably sneak it past mypy:
    # https://github.com/python/mypy/issues/18204#issuecomment-3004743810
    functools.update_wrapper(
        wrapper=wrapper,
        wrapped=func,
    )

    return wrapper


class EventPublisher[P: EventPayload](metaclass=ABCMeta):
    """Interface for event publishers.

    Represents a generic publisher of application metrics events.

    Parameters
    ----------
    application
        Name of the application to include in events.
    event_class
        Fully-enriched class to which payloads will be converted before
        publication.
    """

    def __init__(self, application: str, event_class: type[P]) -> None:
        self._application = application
        self._event_class = event_class

    def construct_event(self, payload: P) -> P:
        """Construct the full event as it will be published.

        Parameters
        ----------
        pyaload
            Payload to publish.

        Returns
        -------
        EventPayload
            Enriched event model including generic metadata and Avro schema
            configuration.
        """
        time_ns = time.time_ns()
        metadata = EventMetadata(
            id=uuid4(),
            application=self._application,
            timestamp=self._ns_to_datetime(time_ns),
            timestamp_ns=time_ns,
        )
        return self._event_class(
            **metadata.model_dump(), **payload.model_dump()
        )

    @abstractmethod
    async def publish(self, payload: P) -> EventMetadata:
        """Publish an event payload.

        Parameters
        ----------
        payload
            Payload to publish.

        Returns
        -------
        EventMetadata
            Full, enriched-with-metadata event model that was published. This
            will be an object of a class derived from both the type of the
            payload and `~safir.metrics.EventMetadata`, but it is typed as the
            latter since that is the newly-added information that may be of
            interest to the caller and therefore the most likely to be
            accessed.
        """

    @staticmethod
    def _ns_to_datetime(ns: int) -> datetime:
        """Convert an `int` number of nanoseconds to a `~datetime.datetime`."""
        return datetime.fromtimestamp(ns / 1e9, tz=UTC)


class KafkaEventPublisher[P: EventPayload](EventPublisher[P]):
    """Publishes one type of event.

    You shouldn't need to instantiate instances of this class yourself.
    Instead, call `~safir.metrics.EventManager.create_publisher`. This object
    wraps a FastStream publisher, schema information, and the underlying event
    manager that will be used to publish events.

    Parameters
    ----------
    application
        Name of the application publishing events.
    manager
        The EventManager that will actually publish the event to Kafka
    event_class
        An ``AvroBaseModel`` with the fields from the payload,
        `~safir.metrics.EventMetadata`, and an inner ``Meta`` class with
        additional schema configuration.
    publisher
        FastStream publisher to use to publish the event to Kafka. This
        contains Kafka metadata, like the topic to publish to.
    schema_info
        Confluent schema information for this event type.
    """

    def __init__(
        self,
        *,
        application: str,
        manager: "KafkaEventManager",
        event_class: type[P],
        publisher: AsyncAPIDefaultPublisher,
        schema_info: SchemaInfo,
    ) -> None:
        super().__init__(application, event_class)
        self._manager = manager
        self._publisher = publisher
        self._schema_info = schema_info

    @override
    async def publish(self, payload: P) -> EventMetadata:
        event = self.construct_event(payload)
        await self._manager.publish(event, self._publisher, self._schema_info)
        return cast("EventMetadata", event)


class NoopEventPublisher[P: EventPayload](EventPublisher[P]):
    """Event publisher that quietly does nothing.

    This is used in applications when event publishing is disabled, so that
    the parsing and conversion of the events is still tested (thus making
    application tests meaningful), but the event is not sent anywhere.
    """

    def __init__(
        self,
        application: str,
        event_class: type[P],
        logger: BoundLogger,
    ) -> None:
        super().__init__(application, event_class)
        self._logger = logger

    @override
    async def publish(self, payload: P) -> EventMetadata:
        event = self.construct_event(payload)
        self._logger.debug(
            "Would have published event", metrics_event=event.model_dump()
        )
        return cast("EventMetadata", event)


class FailedEventPublisher[P: EventPayload](EventPublisher[P]):
    """Event publisher that logs error messages instead of publishing.

    This is used when app metrics initialization has failed due to underlying
    infrastructure issues, but we don't want to crash the instrumented app.
    """

    def __init__(
        self,
        application: str,
        event_class: type[P],
        backoff_interval: timedelta,
        logger: BoundLogger,
    ) -> None:
        super().__init__(application, event_class)
        self.logger = logger
        self._backoff_interval = backoff_interval
        self._publish_called_at = datetime(1900, 1, 1, tzinfo=UTC)

    @override
    async def publish(self, payload: P) -> EventMetadata:
        """Don't do anything except log a message once per backoff interval."""
        now = datetime.now(tz=UTC)
        elapsed = now - self._publish_called_at
        if elapsed > self._backoff_interval:
            self._publish_called_at = now
            self.logger.error(
                "The app metrics system is in an error state. No metrics will"
                " be published by this app. When the underlying infrastructure"
                " (Kafka, Schema Manager, etc) is healthy again, this app must"
                " be restarted.",
                attempted_operation="publish",
            )
        event = self.construct_event(payload)
        return cast("EventMetadata", event)


class MockEventPublisher[P: EventPayload](NoopEventPublisher[P]):
    """Event publisher that quietly does nothing and records all payloads.

    This is meant to be used in unit tests to enable assertions on published
    payloads. It should NOT be used in any deployed application instances
    because memory usage will grow unbounded with each published event.
    """

    def __init__(
        self,
        application: str,
        event_class: type[P],
        logger: BoundLogger,
    ) -> None:
        super().__init__(application, event_class, logger)
        self._published: PublishedList[P] = PublishedList()

    @override
    async def publish(self, payload: P) -> EventMetadata:
        event = await super().publish(payload)
        self._published.append(payload)
        return event

    @property
    def published(self) -> PublishedList[P]:
        """A list of published event payloads with some test helpers."""
        return self._published


class EventManager(metaclass=ABCMeta):
    """Interface for a client for publishing application metrics events.

    This interface is implemented by the fully-functional
    `~safir.metrics.KafkaEventManager` and the disabled
    `~safir.metrics.NoopEventManager`. The latter is used when events
    publishing is disabled so that the rest of the application code doesn't
    need to change.

    Parameters
    ----------
    topic
        Kafka topic to which events will be published.
    logger
        Logger to use. The ``safir.metrics`` logger will be used if none is
        provided.

    Attributes
    ----------
    topic
        Kafka topic to which events will be published.
    logger
        Logger that subclasses should use. This should not be used outside of
        subclasses of this class.
    """

    def __init__(
        self,
        topic: str,
        logger: BoundLogger | None = None,
    ) -> None:
        self.topic = topic
        self.logger = logger or structlog.get_logger("safir.metrics")
        self._publishers: dict[str, EventPublisher] = {}
        self._state = _State.uninitialized

    async def aclose(self) -> None:
        """Shut down any internal state or managed clients."""
        self._publishers = {}
        self._state = _State.uninitialized

    @abstractmethod
    async def build_publisher_for_model[P: EventPayload](
        self, model: type[P]
    ) -> EventPublisher[P]:
        """Implementation-specific construction of the event publisher.

        This class must be overridden by child classes to do the
        implementation-specific work of constructing an appropriate child
        instance of `~safir.metrics.EventPublisher`.

        Parameters
        ----------
        model
            Enriched and configured model representing the event that will be
            published.

        Returns
        -------
        EventPublisher
            An appropriate event publisher implementation instance.
        """

    async def create_publisher[P: EventPayload](
        self, name: str, payload_model: type[P]
    ) -> EventPublisher[P]:
        """Create an `~safir.metrics.EventPublisher` for a type of event.

        The schema is registered with the schema manager when this method is
        called.

        Parameters
        ----------
        name
            Name of the event. This will be used as the name of the Avro
            schema and record and the name of the event in the event storage
            backend.
        payload_model
            A type derived from `~safir.metrics.EventPayload`. This defines
            the type of models that will be passed into the
            `~safir.metrics.EventPublisher.publish` method of the resulting
            publisher. The events as published will include the information in
            this model plus the fields of `~safir.metrics.EventMetadata`.

        Returns
        -------
        EventPublisher
            A publisher for a type of event matching the ``payload_model``.

        Raises
        ------
        DuplicateEventError
            Raised if a publisher with the same name was already registered.
        EventManagerUnintializedError
            Raised if the `initialize` method was not been called before
            calling this method.
        KafkaTopicError
            Raised if the topic for publishing events doesn't exist or we
            don't have access to it.
        """
        if self._state == _State.uninitialized:
            msg = "Initialize EventManager before creating event publishers"
            raise EventManagerUnintializedError(msg)
        if name in self._publishers:
            raise DuplicateEventError(name)

        # Mixin used to configure dataclasses-avroschema.
        class MetaBase(AvroBaseModel):
            class Meta:
                schema_name = name
                namespace = self.topic

        # Construct the event model.
        model = cast(
            "type[P]",
            create_model(
                "EventModel",
                __base__=(payload_model, EventMetadata, MetaBase),
            ),
        )

        # Validate the structure of the model. This verifies that it can be
        # serialized correctly and will raise an exception if it cannot be.
        model.validate_structure()

        # Build the publisher, store it to detect duplicates, and return it.
        publisher = await self.build_publisher_for_model(model)
        self._publishers[name] = publisher
        return publisher

    async def initialize(self) -> None:
        """Initialize any internal state or managed clients.

        This method must be called before calling
        `~safir.metrics.EventManager.create_publisher`.
        """
        self._state = _State.initialized


class KafkaEventManager(EventManager):
    """A tool for publishing application metrics events.

    Events are published to Kafka as avro-serialized messages. The schemas for
    the messages are stored in a Confluent Schema Registry, and this
    EventManager will manage those schemas and throw an exception if the event
    payloads evolve in an incompatible way.

    Parameters
    ----------
    application
        Name of the application that is generating events.
    topic_prefix
        Kafka topic prefix for the metrics events topic for this application.
    kafka_broker
        Broker to use to publish events to Kafka.
    kafka_admin_client
        Admin client to Kafka used to check that it is prepared for event
        publishing. For example, it is used to check if the topic exists.
    schema_manager
        Client to the Confluent-compatible schema registry.
    backoff_interval
        The amount of time to wait until further operations are attempted after
        the KafkaEventManager is put into an error state.
    manage_kafka_broker
        If `True`, start the ``kafka_broker`` on
        `~safir.metrics.EventManager.initialize` and close the
        ``kafka_broker`` when `~safir.metrics.EventManager.aclose` is called.
        If your app's only use of Kafka is to publish metrics events, then
        this should be `True`. If you have a FastStream app that already
        configures a Kafka broker that you want to reuse for metrics, this
        should probably be `False`, and you should pass in your existing Kafka
        broker. In this case, you will need to start the broker before calling
        `~safir.metrics.EventManager.initialize` and stop it after closing the
        event manager.
    raise_on_error
        True if we should raise an exception whenever there is an error with
        the metrics system dependencies, like Kafka or the Schema Manager.
        False if we should just log an error instead. This should be False for
        most production apps so that issues with the metrics infrastructure
        don't bring down the app.
    logger
        Logger to use for internal logging.

    Examples
    --------
    .. code-block:: python

       from safir.kafka import KafkaConnectionSettings, SchemaManagerSettings
       from safir.metrics import (
           EventsConfiguration,
           EventPayload,
           KafkaMetricsConfiguration,
       )


       config = KafkaMetricsConfiguration(
           events=EventsConfiguration(
               application="myapp",
               topic_prefix="what.ever",
           ),
           kafka=KafkaConnectionSettings(
               bootstrap_servers="someserver:1234",
               security_protocol=KafkaSecurityProtocol.PLAINTEXT,
           ),
           schema_manager=SchemaRegistryConnectionSettings(
               registry_url=AnyUrl("https://some.registry")
           ),
       )
       manager = config.make_manager()


       class MyEvent(EventPayload):
           foo: str


       publisher = manager.create_publisher(MyEvent)
       await manager.register_and_initialize()

       await publisher.publish(MyEvent(foo="bar1"))
       await publisher.publish(MyEvent(foo="bar2"))

       await manager.aclose()
    """

    def __init__(
        self,
        *,
        application: str,
        topic_prefix: str,
        kafka_broker: KafkaBroker,
        kafka_admin_client: AIOKafkaAdminClient,
        schema_manager: PydanticSchemaManager,
        backoff_interval: timedelta = EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL,
        manage_kafka_broker: bool = False,
        raise_on_error: bool = False,
        logger: BoundLogger | None = None,
    ) -> None:
        super().__init__(f"{topic_prefix}.{application}", logger)
        self._application = application
        self._broker = kafka_broker
        self._admin_client = kafka_admin_client
        self._schema_manager = schema_manager
        self._backoff_interval = backoff_interval
        self._manage_kafka_broker = manage_kafka_broker
        self._raise_on_error = raise_on_error

    @override
    @abandonable
    async def aclose(self) -> None:
        """Clean up the Kafka clients if they are managed."""
        if self._manage_kafka_broker:
            await self._broker.stop()
        await self._admin_client.close()
        await super().aclose()

    @override
    async def build_publisher_for_model[P: EventPayload](
        self, model: type[P]
    ) -> EventPublisher[P]:
        """Build a Kafka publisher for a specific enriched model.

        Parameters
        ----------
        model
            Enriched and configured model representing the event that will be
            published.

        Returns
        -------
        EventPublisher
            An appropriate event publisher implementation instance. We don't
            want to crash the whole application if we can't build a publisher
            due to problems like the metrics infrastructure being unavailable,
            so return a FailedEventPublisher in those cases.
        """
        publisher = await self._build_publisher_for_model(model)

        if publisher is None:
            return FailedEventPublisher(
                application=self._application,
                event_class=model,
                backoff_interval=self._backoff_interval,
                logger=self.logger,
            )
        else:
            return publisher

    @abandonable
    async def _build_publisher_for_model[P: EventPayload](
        self, model: type[P]
    ) -> EventPublisher[P]:
        async_publisher = self._broker.publisher(self.topic, schema=model)

        # Verify that the topic exists.
        if not await self._is_topic_ok(self.topic):
            raise KafkaTopicError(self.topic)

        # Register the Avro schema if necessary and get the schema details.
        schema_info = await self._schema_manager.register_model(
            model, compatibility=Compatibility.NONE
        )

        # Return the corresponding event publisher.
        return KafkaEventPublisher[P](
            application=self._application,
            event_class=model,
            publisher=async_publisher,
            manager=self,
            schema_info=schema_info,
        )

    @override
    @abandonable
    async def initialize(self) -> None:
        """Initialize the Kafka clients if they are managed."""
        if self._manage_kafka_broker:
            await self._broker.start()
        await self._admin_client.start()
        self._state = _State.initialized

    @abandonable
    async def publish(
        self,
        event: EventPayload,
        publisher: AsyncAPIDefaultPublisher,
        schema_info: SchemaInfo | None,
    ) -> None:
        """Serialize an event to Avro and publish it to Kafka.

        This method should generally not be called directly. It will be called
        from `~safir.metrics.KafkaEventPublisher.publish`.

        Parameters
        ----------
        event
            Fully-enhanced event to publish.
        publisher
            FastStream publisher to use to publish the event.
        schema_info
            Confluent schema registry information about the event type.

        Raises
        ------
        EventManagerUnintializedError
            Raised if the `initialize` method was not been called before
            calling this method.
        """
        if self._state == _State.uninitialized:
            msg = "Initialize EventManager before creating event publishers"
            raise EventManagerUnintializedError(msg)
        encoded = await self._schema_manager.serialize(event)
        await publisher.publish(encoded, no_confirm=True)
        self.logger.debug(
            "Published metrics event",
            metrics_event=event.model_dump(),
            topic=publisher.topic,
            schema_info=schema_info,
        )

    async def _is_topic_ok(self, topic: str) -> bool:
        """Check whether the given Kafka topic is valid.

        Parameters
        ----------
        topic
            Kafka topic to check.

        Returns
        -------
        bool
            `True` if the Kafka topic is valid, `False` otherwise.
        """
        info = await self._admin_client.describe_topics([topic])
        if len(info) == 0:
            return False
        topic_info = info[0]
        error_code = topic_info.get("error_code", None)
        return error_code == 0


class NoopEventManager(EventManager):
    """An event manager that creates publishers that quietly do nothing.

    This is used as the implementation of `~safir.metrics.EventManager` when
    event publication is disabled. The event type registrations and event
    payloads are still verified to catch errors, but are then discarded.


    Parameters
    ----------
    application
        Name of the application that is generating events.
    topic_prefix
        Kafka topic prefix for the metrics events topic for this application.
    logger
        Logger to use for internal logging.
    """

    def __init__(
        self,
        application: str,
        topic_prefix: str,
        logger: BoundLogger | None = None,
    ) -> None:
        super().__init__(f"{topic_prefix}.{application}", logger)
        self._application = application

    @override
    async def build_publisher_for_model[P: EventPayload](
        self, model: type[P]
    ) -> EventPublisher[P]:
        """Build a no-op publisher for a specific enriched model.

        Parameters
        ----------
        model
            Enriched and configured model representing the event that will be
            published.

        Returns
        -------
        EventPublisher
            An appropriate event publisher implementation instance.
        """
        return NoopEventPublisher[P](self._application, model, self.logger)


class MockEventManager(EventManager):
    """An event manager that creates mock publishers that record all publishes.

    This is used as the implementation of `~safir.metrics.EventManager` when
    event publication is disabled and mocking is enabled. Like a
    `~safir.metrics.NoopEventManager`, the event type registrations are still
    verified to catch errors, but any calls to
    `~safir.metrics.MockEventPublisher` are recorded for later assertion.

    This is for use only in unit testing. Don't use it in any deployed
    environment because memory usage will grow unbounded.

    Parameters
    ----------
    application
        Name of the application that is generating events.
    topic_prefix
        Kafka topic prefix for the metrics events topic for this application.
    logger
        Logger to use for internal logging.
    """

    def __init__(
        self,
        application: str,
        topic_prefix: str,
        logger: BoundLogger | None = None,
    ) -> None:
        super().__init__(f"{topic_prefix}.{application}", logger)
        self._application = application

    @override
    async def build_publisher_for_model[P: EventPayload](
        self, model: type[P]
    ) -> EventPublisher[P]:
        """Build a no-op recording publisher for a specific enriched model.

        Parameters
        ----------
        model
            Enriched and configured model representing the event that will be
            published.

        Returns
        -------
        EventPublisher
            An appropriate event publisher implementation instance.
        """
        return MockEventPublisher[P](self._application, model, self.logger)



"""Support functions for testing UWS code."""

from __future__ import annotations

from datetime import timedelta
from typing import Annotated

from arq.connections import RedisSettings
from fastapi import Form, Query
from pydantic import BaseModel, SecretStr
from vo_models.uws import JobSummary, Parameter, Parameters

from safir.arq import ArqMode
from safir.uws import ParametersModel, UWSConfig, UWSRoute

__all__ = [
    "SimpleParameters",
    "SimpleXmlParameters",
    "build_uws_config",
]


class SimpleXmlParameters(Parameters):
    name: Parameter = Parameter(id="name")


class SimpleWorkerParameters(BaseModel):
    name: str


class SimpleParameters(
    ParametersModel[SimpleWorkerParameters, SimpleXmlParameters]
):
    name: str

    def to_worker_parameters(self) -> SimpleWorkerParameters:
        return SimpleWorkerParameters(name=self.name)

    def to_xml_model(self) -> SimpleXmlParameters:
        return SimpleXmlParameters(name=Parameter(id="name", value=self.name))


async def _get_dependency(
    name: Annotated[str, Query()],
) -> SimpleParameters:
    return SimpleParameters(name=name)


async def _post_dependency(
    name: Annotated[str, Form()],
) -> SimpleParameters:
    return SimpleParameters(name=name)


def build_uws_config() -> UWSConfig:
    """Set up a test configuration."""
    return UWSConfig(
        arq_mode=ArqMode.test,
        arq_redis_settings=RedisSettings(host="localhost", port=6379),
        async_post_route=UWSRoute(
            dependency=_post_dependency, summary="Create async job"
        ),
        execution_duration=timedelta(minutes=10),
        job_summary_type=JobSummary[SimpleXmlParameters],
        lifetime=timedelta(days=1),
        parameters_type=SimpleParameters,
        signing_service_account="signer@example.com",
        slack_webhook=SecretStr("https://example.com/fake-webhook"),
        sync_get_route=UWSRoute(
            dependency=_get_dependency, summary="Sync request"
        ),
        sync_post_route=UWSRoute(
            dependency=_post_dependency, summary="Sync request"
        ),
        wobbly_url="https://example.com/wobbly",
        worker="hello",
    )



"""Tests for the safir.metadata module."""

from __future__ import annotations

from email.message import Message
from importlib.metadata import metadata
from typing import cast

import pytest

from safir.metadata import get_metadata, get_project_url


@pytest.fixture(scope="session")
def safir_metadata() -> Message:
    return cast("Message", metadata("safir"))


def test_get_project_url(safir_metadata: Message) -> None:
    """Test the get_project_url function using Safir's own metadata."""
    source_url = get_project_url(safir_metadata, "Source")
    assert source_url == "https://github.com/lsst-sqre/safir"


def test_get_project_url_missing(safir_metadata: Message) -> None:
    """Test that get_project_url returns None for a missing URL."""
    source_url = get_project_url(safir_metadata, "Nonexistent")
    assert source_url is None


@pytest.mark.asyncio
async def test_get_metadata() -> None:
    """Test get_metadata in normal usage."""
    data = get_metadata(package_name="safir", application_name="testapp")
    assert data.name == "testapp"
    assert data.version
    assert data.description
    assert data.repository_url == "https://github.com/lsst-sqre/safir"
    assert data.documentation_url == "https://safir.lsst.io"



"""Tests for the database session dependency."""

from __future__ import annotations

from typing import Annotated

import pytest
import structlog
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient
from sqlalchemy import String, select
from sqlalchemy.ext.asyncio import async_scoped_session
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from safir.database import (
    create_async_session,
    create_database_engine,
    initialize_database,
)
from safir.dependencies.db_session import db_session_dependency


class Base(DeclarativeBase):
    """Declarative base for testing."""


class User(Base):
    """Tiny database table for testing."""

    __tablename__ = "user"

    username: Mapped[str] = mapped_column(String(64), primary_key=True)


@pytest.mark.asyncio
async def test_session(database_url: str, database_password: str) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(engine, logger, schema=Base.metadata, reset=True)
    session = await create_async_session(engine, logger)
    await db_session_dependency.initialize(database_url, database_password)

    app = FastAPI()

    @app.post("/add")
    async def add(
        session: Annotated[
            async_scoped_session, Depends(db_session_dependency)
        ],
    ) -> None:
        async with session.begin():
            session.add(User(username="foo"))

    @app.get("/list")
    async def get_list(
        session: Annotated[
            async_scoped_session, Depends(db_session_dependency)
        ],
    ) -> list[str]:
        async with session.begin():
            result = await session.scalars(select(User.username))
            return list(result.all())

    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/list")
        assert r.status_code == 200
        assert r.json() == []

        r = await client.post("/add")
        assert r.status_code == 200

        r = await client.get("/list")
        assert r.status_code == 200
        assert r.json() == ["foo"]

    # Retrieve the database contents through an entirely separate connection
    # pool to check the transaction was really committed.
    async with session.begin():
        result = await session.scalars(select(User.username))
        assert result.all() == ["foo"]

    await session.remove()
    await engine.dispose()
    await db_session_dependency.aclose()



"""Kafka test container with listeners for different security protocols.

This test container is heavily inspired by the default Testcontainers_ `Kafka
test container
<https://github.com/testcontainers/testcontainers-python/blob/main/modules/kafka/testcontainers/kafka/__init__.py>`__,
but adds additional listeners for different security protocols. The key
changes are:

* Create SSL, SASL_SSL, and SASL_PLAINTEXT listeners in addition to the
  the PLAINTEXT listener.
* Expose the SASL username and password with ``get_sasl_username`` and
  ``get_sasl_password`` methods.
* Provide a ``get_secret_file_contents`` method to retrieve the TLS
  certificates created during startup.
* Provide a ``reset`` method to delete all topics from the running Kafka
  instance.
* Always limit the broker to the first host.
* Always use KRaft.
"""

import errno
import os
import sys
import tarfile
import time
from importlib import resources
from io import BytesIO
from pathlib import Path
from random import randint
from textwrap import dedent
from typing import Any, Self

from testcontainers.core.container import DockerContainer
from testcontainers.core.version import ComparableVersion
from testcontainers.core.waiting_utils import wait_for_logs

from ._constants import CONFLUENT_VERSION_TAG, KAFKA_REPOSITORY

__all__ = ["FullKafkaContainer"]


class FullKafkaContainer(DockerContainer):
    """Kafka container.

    Examples
    --------

    .. code-block:: python

       from testcontainers.kafka import KafkaContainer

       with KafkaContainer() as kafka:
           connection = kafka.get_bootstrap_server()
    """

    _CLUSTER_ID = "MkU3OEVBNTcwNTJENDM2Qk"
    """Kafka cluster ID.

    This is arbitrary and has no special meaning. The cluster just needs some
    cluster ID and this is the one used as an example in some Kafka
    documentation.
    """

    _MIN_KRAFT_TAG = "7.0.0"
    """Minimum tag version for KRaft support."""

    _START_SCRIPT = "/tc-start.sh"
    """Path where the start script is stored in the container."""

    def __init__(
        self,
        image: str = f"{KAFKA_REPOSITORY}:{CONFLUENT_VERSION_TAG}",
        port: int = 9093,
        ssl_port: int = 29093,
        sasl_plaintext_port: int = 29094,
        sasl_ssl_port: int = 29095,
        *,
        constant_host_ports: bool = False,
        **kwargs: Any,
    ) -> None:
        """Start a Kafka TestContainer.

        Parameters
        ----------
        image
            The Kafka Docker image to use
        port
            Kafka plaintext auth port
        ssl_port
            Kafka SSL auth port
        sasl_plaintext_port
            Kafka SASL plaintext auth port
        sasl_ssl_port
            Kafka SASL SSL auth port
        constant_host_ports
            If True, the same host ports will be mapped every time. This is
            useful when stopping and starting Kafka during a test. Be careful,
            because if the test fails, then you won't be able to use this
            TestContainer again until the previous container is removed because
            there will be a port conflic.
        """
        self._verify_min_kraft_version(image)
        super().__init__(image, **kwargs)

        self.port = port
        self.sasl_plaintext_port = sasl_plaintext_port
        self.ssl_port = ssl_port
        self.sasl_ssl_port = sasl_ssl_port

        starting_host_port = randint(51234, 54321)
        for offset, container_port in enumerate(
            (
                self.port,
                self.sasl_plaintext_port,
                self.ssl_port,
                self.sasl_ssl_port,
            )
        ):
            if constant_host_ports:
                host_port = starting_host_port + offset
            else:
                host_port = None
            self.with_bind_ports(container_port, host_port)

        self.container_cert_path = Path("/etc/kafka/secrets")
        self.sasl_username = "admin"
        self.sasl_password = "admin"

    def get_bootstrap_server(self) -> str:
        """Get the bootstrap server for a no TLS, no SASL connection."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        return f"{host}:{port}"

    def get_ssl_bootstrap_server(self) -> str:
        """Get the bootstrap server for a TLS connection."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.ssl_port)
        return f"{host}:{port}"

    def get_sasl_ssl_bootstrap_server(self) -> str:
        """Get the bootstrap server for a SASL TLS connection."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.sasl_ssl_port)
        return f"{host}:{port}"

    def get_sasl_plaintext_bootstrap_server(self) -> str:
        """Get the bootstrap server for a SASL PLAINTEXT connection."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.sasl_plaintext_port)
        return f"{host}:{port}"

    def get_sasl_username(self) -> str:
        """Get the configured SASL username for the Kafka listener."""
        return self.sasl_username

    def get_sasl_password(self) -> str:
        """Get the configured SASL password for the Kafka listener."""
        return self.sasl_password

    def start(self, timeout: int = 30) -> Self:
        """Start the container.

        Parameters
        ----------
        timeout
            How long to wait for the container to start before raising an
            error if Kafka still isn't running.
        """
        self._configure()

        # Create a start command that waits for the start script to be created
        # and then runs it. We have to do it this way because the start script
        # is not baked into the container and we can't add it to the
        # container until the container is started.
        script = self._START_SCRIPT
        command = (
            f'sh -c "while [ ! -f {script} ]; do sleep 0.1; done; sh {script}"'
        )
        self.with_command(command)

        # Start the container and then add the startup script, which should be
        # picked up and run.
        super().start()
        self._create_start_script()
        self.wait_for_ready(timeout)

        # Configure SASL.
        self.exec(
            f"kafka-configs --bootstrap-server localhost:9092 --alter "
            f"--add-config 'SCRAM-SHA-512=[iterations=8192,"
            f"password={self.sasl_password}]' --entity-type users "
            f"--entity-name {self.sasl_username}"
        )

        return self

    def wait_for_ready(self, timeout: int = 30) -> None:
        """Wait until Kafka is ready to serve requests."""
        try:
            wait_for_logs(self, r".*Kafka Server started.*", timeout=timeout)
        except TimeoutError:
            output, errors = self.get_logs()
            sys.stderr.write(errors.decode())
            sys.stdout.write(output.decode())
            raise

    def wait_for_ready_again(self, timeout: int = 30) -> None:
        """Wait until Kafka is ready to serve requests."""
        try:
            wait_for_logs(
                self,
                r"Kafka Server started.*\n(?:.*\n)+.*Kafka Server started",
                timeout=timeout,
            )
        except TimeoutError:
            output, errors = self.get_logs()
            sys.stderr.write(errors.decode())
            sys.stdout.write(output.decode())
            raise

    def stop_container(self) -> None:
        """Stop the container, but don't remove it.

        This is useful in tests for stopping the container and starting it up
        again with the same ports mapped. The default stop method removes the
        container too, so if you call start again, it hass different ports
        mapped.
        """
        container = self.get_wrapped_container()
        if not container:
            raise RuntimeError("Container does not exist.")
        container.stop()

    def start_container_again(self) -> None:
        """Start this container again.

        Can only be called after the container is stopped with stop_container.
        """
        container = self.get_wrapped_container()
        if not container:
            raise RuntimeError("Container does not exist.")
        container.start()
        self.wait_for_ready_again()

    def create_file(self, content: bytes, path: str) -> None:
        """Create a file inside the container.

        Parameters
        ----------
        content
            Content of the file.
        path
            Path to the file inside the container.
        """
        with BytesIO() as archive:
            with tarfile.TarFile(fileobj=archive, mode="w") as tar:
                tarinfo = tarfile.TarInfo(name=path)
                tarinfo.size = len(content)
                tarinfo.mtime = int(time.time())
                tar.addfile(tarinfo, BytesIO(content))
                archive.seek(0)
                self.get_wrapped_container().put_archive("/", archive)

    def get_secret_file_contents(self, name: str) -> str:
        """Get the contents of a file from :file:`/etc/kafka/secrets`.

        Useful for getting TLS certs and keys for connecting with the SSL
        security protocol.

        We can't just mount a host volume over :file:`/etc/kafka/secrets`
        because the OS user that owns that directory could be different than
        that ``appuser`` user that runs kafka in the container.

        Parameters
        ----------
        name
            Name of the secret to retrieve. This should normally be one of
            :file:`ca.crt`, :file:`client.crt`, or :file:`client.key`.

        Raises
        ------
        FileNotFoundError
            Raised if the requested file could not be found in the
            :file:`/etc/kafka/secrets` directory.
        """
        path = self.container_cert_path / name
        bits, _ = self.get_wrapped_container().get_archive(str(path))
        with BytesIO() as raw:
            for chunk in bits:
                raw.write(chunk)
            raw.seek(0)
            with tarfile.open(fileobj=raw) as tf:
                stream = tf.extractfile(name)
                if not stream:
                    raise FileNotFoundError(
                        errno.ENOENT, os.strerror(errno.ENOENT), str(path)
                    )
                return stream.read().decode()

    def reset(self) -> None:
        """Reset all Kafka topics."""
        self.exec(
            "/bin/kafka-topics --bootstrap-server localhost:9092 --delete"
            " --topic '.*'"
        )

    def _configure(self) -> None:
        """Configure the container environment variables."""
        quorum_voters = f"1@{self._get_network_alias()}:9094"

        # Kafka
        self.with_env("CLUSTER_ID", self._CLUSTER_ID)
        self.with_env("KAFKA_BROKER_ID", "1")
        self.with_env("KAFKA_CONTROLLER_LISTENER_NAMES", "CONTROLLER")
        self.with_env("KAFKA_CONTROLLER_QUORUM_VOTERS", quorum_voters)
        self.with_env("KAFKA_GROUP_INITIAL_REBALANCE_DELAY_MS", "0")
        self.with_env(
            "KAFKA_LISTENER_SECURITY_PROTOCOL_MAP",
            (
                "BROKER:PLAINTEXT,PLAINTEXT:PLAINTEXT,SSL:SSL,SASL_SSL:"
                "SASL_SSL,SASL_PLAINTEXT:SASL_PLAINTEXT,CONTROLLER:PLAINTEXT"
            ),
        )
        self.with_env("KAFKA_INTER_BROKER_LISTENER_NAME", "BROKER")
        self.with_env(
            "KAFKA_LISTENERS",
            (
                f"PLAINTEXT://0.0.0.0:{self.port},"
                "BROKER://0.0.0.0:9092,"
                f"SSL://0.0.0.0:{self.ssl_port},"
                f"SASL_SSL://0.0.0.0:{self.sasl_ssl_port},"
                f"SASL_PLAINTEXT://0.0.0.0:{self.sasl_plaintext_port},"
                "CONTROLLER://0.0.0.0:9094"
            ),
        )
        self.with_env("KAFKA_LOG_FLUSH_INTERVAL_MESSAGES", "10000000")
        self.with_env("KAFKA_NODE_ID", 1)
        self.with_env("KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR", "1")
        self.with_env("KAFKA_OFFSETS_TOPIC_NUM_PARTITIONS", "1")
        self.with_env("KAFKA_PROCESS_ROLES", "broker,controller")

        # SSL
        self.with_env("KAFKA_SSL_KEYSTORE_FILENAME", "server.keystore.jks")
        self.with_env("KAFKA_SSL_KEYSTORE_CREDENTIALS", "credentials")
        self.with_env("KAFKA_SSL_KEY_CREDENTIALS", "credentials")
        self.with_env("KAFKA_SSL_TRUSTSTORE_FILENAME", "server.truststore.jks")
        self.with_env("KAFKA_SSL_TRUSTSTORE_CREDENTIALS", "credentials")

        # SASL
        self.with_env("KAFKA_SASL_ENABLED_MECHANISMS", "SCRAM-SHA-512")
        self.with_env(
            "KAFKA_OPTS",
            (
                "-Djava.security.auth.login.config="
                f"{self.container_cert_path!s}/kafka_server_jaas.conf"
            ),
        )

    def _create_start_script(self) -> None:
        """Create the script to start the container."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        ssl_port = self.get_exposed_port(self.ssl_port)
        sasl_ssl_port = self.get_exposed_port(self.sasl_ssl_port)
        sasl_plaintext_port = self.get_exposed_port(self.sasl_plaintext_port)
        listeners = (
            f"PLAINTEXT://{host}:{port},"
            f"BROKER://$(hostname -i | cut -d' ' -f1):9092,"
            f"SSL://{host}:{ssl_port},SASL_SSL://{host}:{sasl_ssl_port},"
            f"SASL_PLAINTEXT://{host}:{sasl_plaintext_port}"
        )

        files = resources.files("safir.testing.containers").joinpath("data")
        with files.joinpath("generate-kafka-secrets.bash").open("r") as f:
            generate_certs_script = f.read().encode()
        self.create_file(generate_certs_script, "/generate-kafka-secrets.bash")

        data = (
            dedent(
                f"""
                #!/bin/bash
                /bin/bash /generate-kafka-secrets.bash \
                    {self.container_cert_path} {host}
                sed -i '/KAFKA_ZOOKEEPER_CONNECT/d' \
                    /etc/confluent/docker/configure
                echo 'kafka-storage format --ignore-formatted \
                  -t {self._CLUSTER_ID} -c /etc/kafka/kafka.properties' \
                  >> /etc/confluent/docker/configure
                export KAFKA_ADVERTISED_LISTENERS={listeners}
                . /etc/confluent/docker/bash-config
                /etc/confluent/docker/configure
                /etc/confluent/docker/launch
                """
            )
            .strip()
            .encode()
        )
        self.create_file(data, self._START_SCRIPT)

    def _get_network_alias(self) -> str | None:
        """Get the alias of the network, if any."""
        if self._network:
            fallback = [self._network.name or self._kwargs.get("network", [])]
            return next(iter(self._network_aliases or fallback), None)
        return "localhost"

    def _verify_min_kraft_version(self, image: str) -> None:
        """Verify that the image is new enough to suppot KRaft.

        Parameters
        ----------
        image
            Docker image to run.
        """
        actual_version = image.split(":")[-1]
        if ComparableVersion(actual_version) < self._MIN_KRAFT_TAG:
            raise ValueError(
                f"Provided Confluent Platform's version {actual_version} "
                f"is not supported (must be {self._MIN_KRAFT_TAG} or above)"
            )



"""Error handling support for FastAPI."""

from __future__ import annotations

from typing import ClassVar

from fastapi import Request, status
from fastapi.responses import JSONResponse

from safir.models import ErrorLocation
from safir.slack.webhook import SlackIgnoredException

__all__ = [
    "ClientRequestError",
    "client_request_error_handler",
]


class ClientRequestError(SlackIgnoredException):
    """Represents an error in a client request.

    This is a base class for exceptions representing expected errors in client
    requests. Expected here means that the error should result in a 4xx HTTP
    error code with an error message in the body and do not represent a server
    failure that should produce alerts. It should normally be used in
    conjunction with the `client_request_error_handler` exception handler.

    Exceptions inheriting from this class should set the class variable
    ``error`` to a unique error code (normally composed of lowercase letters
    and underscores) for that error, and the class variable ``status_code`` to
    the HTTP status code this exception should generate (if not 422, the
    default).

    Attributes
    ----------
    location
        The part of the request giving rise to the error. This can be set by
        catching the exception in the part of the code that knows where the
        data came from, setting this attribute, and re-raising the exception.
    field_path
        Field, as a hierarchical list of structure elements, within that part
        of the request giving rise to the error. As with ``location``, can be
        set by catching and re-raising.

    Parameters
    ----------
    message
        Error message, used as the ``msg`` key in the serialized error.
    location
        The part of the request giving rise to the error. This may be omitted
        if the error message does not have meaningful location information, or
        to set this information via the corresponding attribute in a later
        exception handler.
    field_path
        Field, as a hierarchical list of structure elements, within the
        ``location`` giving rise to the error. This may be omitted if the
        error message does not have meaningful location information, or to set
        this information via the corresponding attribute in a later exception
        handler.

    Examples
    --------
    This class is meant to be subclassed. For example:

    .. code-block:: python

       from fastapi import status
       from safir.fastapi import ClientRequestError


       class UnknownUserError(ClientRequestError):
           error = "unknown_user"
           status_code = status.HTTP_404_NOT_FOUND

    If the location of the error is known when it is raised, it can be passed
    to the constructor. If a given error message always stems from the same
    location information, override ``__init__`` of the corresponding child
    exception class to pass that location information to the parent
    constructor.

    More commonly, the location information is only known to the handler, but
    the exception will be raised by internal service code. In this case, do
    not specify ``location`` or ``field_path`` when raising the exception, and
    catch and re-raise the exception in the handler after adding additional
    location information. For instance:

    .. code-block:: python

       from safir.models import ErrorLocation


       @router.get("/info/{username}", response_model=UserInfo)
       async def get_info(username: str) -> UserInfo:
           try:
               return get_user_data(username)
           except UnknownUserError as e:
               e.location = ErrorLocation.path
               e.field_path = ["username"]
               raise

    ``field_path`` may be a hierarchical list of elements in a complex
    structure. For example, for a POST handler that accepts JSON and finds a
    problem in the ``address`` key of the ``user`` field, set ``field_path``
    to ``["user", "address"]``.

    Some errors, such as permission denied errors, will not have any
    meaningful associated location information. In this case, do not specify
    ``location`` or ``field_path``.

    Notes
    -----
    FastAPI_ uses a standard error model for internally-generated errors
    (mostly Pydantic_ validation errors), but doesn't provide a facility to
    generate errors in the same format. Safir provides the
    `~safir.models.ErrorModel` Pydantic model, this class, and the
    `client_request_error_handler` exception handler to fill in the gap.

    This class records the necessary information to generate a
    FastAPI-compatible error structure and knows how to serialize itself in
    the same structure as `~safir.models.ErrorModel` via the ``to_dict``
    method. The `client_request_error_handler` exception handler uses that
    facility to transform this exception into a
    `fastapi.responses.JSONResponse` with an appropriate status code.

    The FastAPI error serialization format supports returning multiple errors
    at a time as a list in the ``detail`` key. This functionality is not
    supported by this exception class.
    """

    error: ClassVar[str] = "validation_failed"
    """Used as the ``type`` field of the error message.

    Should be overridden by any subclass.
    """

    status_code: ClassVar[int] = status.HTTP_422_UNPROCESSABLE_ENTITY
    """HTTP status code for this type of validation error."""

    def __init__(
        self,
        message: str,
        location: ErrorLocation | None = None,
        field_path: list[str] | None = None,
    ) -> None:
        super().__init__(message)
        self.location = location
        self.field_path = field_path

    def to_dict(self) -> dict[str, list[str] | str]:
        """Convert the exception to a dictionary suitable for the response.

        Returns
        -------
        dict
            Serialized error message in a format suitable as a member of the
            list passed to the ``detail`` parameter to a
            `fastapi.HTTPException`. It is designed to produce the same JSON
            structure as native FastAPI errors.

        Notes
        -----
        The format of the returned dictionary is the same as the serialization
        of `~safir.models.ErrorDetail`, and is meant to be one element in the
        list that is the value of the ``detail`` key.
        """
        result: dict[str, list[str] | str] = {
            "msg": str(self),
            "type": self.error,
        }
        if self.location:
            if self.field_path:
                result["loc"] = [self.location.value, *self.field_path]
            else:
                result["loc"] = [self.location.value]
        return result


async def client_request_error_handler(
    request: Request, exc: ClientRequestError
) -> JSONResponse:
    """Exception handler for exceptions derived from `ClientRequestError`.

    Parameters
    ----------
    request
        Request that gave rise to the exception.
    exc
        Exception.

    Returns
    -------
    fastapi.JSONResponse
        Serialization of the exception following `~safir.models.ErrorModel`,
        which is compatible with the serialization format used internally by
        FastAPI.

    Examples
    --------
    This function should be installed as a global exception handler for a
    FastAPI app:

    .. code-block:: python

       from safir.fastapi import (
           ClientRequestError,
           client_request_error_handler,
       )

       app.exception_handler(ClientRequestError)(client_request_error_handler)
    """
    return JSONResponse(
        status_code=exc.status_code, content={"detail": [exc.to_dict()]}
    )



"""pytest fixtures for UWS testing."""

from __future__ import annotations

from collections.abc import AsyncGenerator, Iterator
from contextlib import asynccontextmanager
from datetime import timedelta

import pytest
import pytest_asyncio
import respx
import structlog
from asgi_lifespan import LifespanManager
from fastapi import APIRouter, FastAPI
from httpx import ASGITransport, AsyncClient
from structlog.stdlib import BoundLogger

from safir.arq import MockArqQueue
from safir.logging import configure_logging
from safir.middleware.x_forwarded import XForwardedMiddleware
from safir.slack.webhook import SlackRouteErrorHandler
from safir.testing.gcs import MockStorageClient, patch_google_storage
from safir.testing.slack import MockSlackWebhook, mock_slack_webhook
from safir.testing.uws import MockUWSJobRunner, MockWobbly, patch_wobbly
from safir.uws import UWSApplication, UWSConfig
from safir.uws._dependencies import UWSFactory, uws_dependency

from ..support.uws import build_uws_config


@pytest_asyncio.fixture
async def app(
    arq_queue: MockArqQueue,
    mock_wobbly: MockWobbly,
    uws_config: UWSConfig,
    logger: BoundLogger,
) -> AsyncGenerator[FastAPI]:
    """Return a configured test application for UWS.

    This is a stand-alone test application independent of any real web
    application so that the UWS routes can be tested without reference to
    the pieces added by an application.
    """
    uws = UWSApplication(uws_config)
    uws.override_arq_queue(arq_queue)

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> AsyncGenerator[None]:
        await uws.initialize_fastapi()
        yield
        await uws.shutdown_fastapi()

    app = FastAPI(lifespan=lifespan)
    app.add_middleware(XForwardedMiddleware)
    uws.install_middleware(app)
    router = APIRouter(route_class=SlackRouteErrorHandler)
    uws.install_handlers(router)
    app.include_router(router, prefix="/test")
    uws.install_error_handlers(app)

    async with LifespanManager(app):
        yield app


@pytest.fixture
def arq_queue() -> MockArqQueue:
    return MockArqQueue()


@pytest_asyncio.fixture
async def client(
    app: FastAPI, test_token: str, test_username: str
) -> AsyncGenerator[AsyncClient]:
    """Return an ``httpx.AsyncClient`` configured to talk to the test app."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="https://example.com/",
        headers={
            "X-Auth-Request-Token": test_token,
            "X-Auth-Request-User": test_username,
        },
    ) as client:
        yield client


@pytest.fixture
def logger() -> BoundLogger:
    configure_logging(name="uws")
    return structlog.get_logger("uws")


@pytest.fixture(autouse=True)
def mock_google_storage() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        expected_expiration=timedelta(minutes=15), bucket_name="some-bucket"
    )


@pytest.fixture(autouse=True)
def mock_slack(
    uws_config: UWSConfig, respx_mock: respx.Router
) -> MockSlackWebhook:
    assert uws_config.slack_webhook
    return mock_slack_webhook(
        uws_config.slack_webhook.get_secret_value(), respx_mock
    )


@pytest.fixture
def mock_wobbly(respx_mock: respx.Router, uws_config: UWSConfig) -> MockWobbly:
    return patch_wobbly(respx_mock, str(uws_config.wobbly_url))


@pytest_asyncio.fixture
async def runner(
    uws_config: UWSConfig, arq_queue: MockArqQueue
) -> MockUWSJobRunner:
    return MockUWSJobRunner(uws_config, arq_queue)


@pytest.fixture
def test_service() -> str:
    return "test-service"


@pytest.fixture
def test_token(test_service: str, test_username: str) -> str:
    return MockWobbly.make_token(test_service, test_username)


@pytest.fixture
def test_username() -> str:
    return "test-user"


@pytest.fixture
def uws_config() -> UWSConfig:
    return build_uws_config()


@pytest_asyncio.fixture
async def uws_factory(app: FastAPI, logger: BoundLogger) -> UWSFactory:
    return await uws_dependency(AsyncClient(), logger)



"""Standard models for FastAPI applications.

Notes
-----
FastAPI does not appear to export its error response model in a usable form,
so define a copy of it so that we can reference it in API definitions to
generate good documentation.
"""

from ._errors import ErrorDetail, ErrorLocation, ErrorModel

__all__ = [
    "ErrorDetail",
    "ErrorLocation",
    "ErrorModel",
]



"""Utility functions for asyncio code."""

from ._multiqueue import AsyncMultiQueue, AsyncMultiQueueError
from ._run import run_with_asyncio

__all__ = [
    "AsyncMultiQueue",
    "AsyncMultiQueueError",
    "run_with_asyncio",
]



"""Managing database engines and sessions."""

from __future__ import annotations

import asyncio
from typing import Any
from urllib.parse import quote, urlparse

from pydantic import SecretStr
from pydantic_core import Url
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    async_scoped_session,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.sql.expression import Select
from structlog.stdlib import BoundLogger

__all__ = [
    "build_database_url",
    "create_async_session",
    "create_database_engine",
]


def build_database_url(
    url: str | Url, password: str | SecretStr | None, *, is_async: bool = True
) -> str:
    """Build the authenticated URL for the database.

    Unless ``is_async`` is set to `False`, the database scheme is forced to
    ``postgresql+asyncpg`` if it is ``postgresql``, and ``mysql+asyncmy`` if
    it is ``mysql``.

    Parameters
    ----------
    url
        Database connection URL, not including the password.
    password
        Database connection password.
    is_async
        Whether to force an async driver.

    Returns
    -------
    url
        URL including the password.

    Raises
    ------
    ValueError
        Raised if a password was provided but the connection URL has no
        username.
    """
    if not isinstance(url, str):
        url = str(url)
    parsed_url = urlparse(url)
    if parsed_url.scheme == "postgresql":
        parsed_url = parsed_url._replace(scheme="postgresql+asyncpg")
    elif parsed_url.scheme == "mysql":
        parsed_url = parsed_url._replace(scheme="mysql+asyncmy")
    if password:
        if isinstance(password, SecretStr):
            password = password.get_secret_value()
        if not parsed_url.username:
            raise ValueError(f"No username in database URL {url}")
        password = quote(password, safe="")

        # The username portion of the parsed URL does not appear to decode URL
        # escaping of the username, so we should not quote it again or we will
        # get double-quoting.
        netloc = f"{parsed_url.username}:{password}@{parsed_url.hostname}"
        if parsed_url.port:
            netloc = f"{netloc}:{parsed_url.port}"
        parsed_url = parsed_url._replace(netloc=netloc)
    return parsed_url.geturl()


def create_database_engine(
    url: str | Url,
    password: str | SecretStr | None,
    *,
    connect_args: dict[str, Any] | None = None,
    isolation_level: str | None = None,
    max_overflow: int | None = None,
    pool_size: int | None = None,
    pool_timeout: float | None = None,
) -> AsyncEngine:
    """Create a new async database engine.

    Parameters
    ----------
    url
        Database connection URL, not including the password.
    password
        Database connection password.
    connect_args
        Additional connection arguments to pass directly to the underlying
        database driver.
    isolation_level
        If specified, sets a non-default isolation level for the database
        engine.
    max_overflow
        Maximum number of connections over the pool size for surge traffic.
    pool_size
        Connection pool size.
    pool_timeout
        How long to wait for a connection from the connection pool before
        giving up.

    Returns
    -------
    sqlalchemy.ext.asyncio.AsyncEngine
        Newly-created database engine.  When done with the engine, the caller
        must call ``await engine.dispose()``.

    Raises
    ------
    ValueError
        A password was provided but the connection URL has no username.
    """
    url = build_database_url(url, password)
    kwargs: dict[str, Any] = {}
    if connect_args:
        kwargs["connect_args"] = connect_args
    if isolation_level:
        kwargs["isolation_level"] = isolation_level
    if max_overflow is not None:
        kwargs["max_overflow"] = max_overflow
    if pool_size is not None:
        kwargs["pool_size"] = pool_size
    if pool_timeout is not None:
        kwargs["pool_timeout"] = pool_timeout
    return create_async_engine(url, **kwargs)


async def create_async_session(
    engine: AsyncEngine,
    logger: BoundLogger | None = None,
    *,
    statement: Select | None = None,
) -> async_scoped_session:
    """Create a new async database session.

    Optionally checks that the database is available and retries in a loop for
    10s if it is not.  This should be used during application startup to wait
    for any network setup or database proxy sidecar.

    Parameters
    ----------
    engine
        Database engine to use for the session.
    logger
        Logger for reporting errors.  Used only if a statement is provided.
    statement
        If provided, statement to run to check database connectivity.  This
        will be modified with ``limit(1)`` before execution.  If not provided,
        database connectivity will not be checked.

    Returns
    -------
    sqlalchemy.ext.asyncio.async_scoped_session
        The database session proxy.  This is an asyncio scoped session that is
        scoped to the current task, which means that it will materialize new
        AsyncSession objects for each asyncio task (and thus each web
        request).  ``await session.remove()`` should be called when the caller
        is done with the session.
    """
    factory = async_sessionmaker(engine, expire_on_commit=False)
    session = async_scoped_session(factory, scopefunc=asyncio.current_task)

    # If no statement was provided, just return the async_scoped_session.
    if statement is None:
        return session

    # A statement was provided, so we want to check connectivity and retry for
    # up to ten seconds before returning the session.
    for _ in range(5):
        try:
            async with session.begin():
                await session.execute(statement.limit(1))
                return session
        except (ConnectionRefusedError, OperationalError, OSError):
            if logger:
                logger.info("database not ready, waiting two seconds")
            await session.remove()
            await asyncio.sleep(2)
            continue

    # If we got here, we failed five times.  Try one last time without
    # catching exceptions so that we raise the appropriate exception to our
    # caller.
    async with session.begin():
        await session.execute(statement.limit(1))
        return session



"""Configuration for publishing events."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Annotated, Any, override

import structlog
from aiokafka.admin.client import AIOKafkaAdminClient
from faststream.kafka import KafkaBroker
from pydantic import AfterValidator, AliasChoices, Field, ValidationError
from pydantic.alias_generators import to_camel
from pydantic_settings import BaseSettings, SettingsConfigDict
from structlog.stdlib import BoundLogger

from safir.pydantic import HumanTimedelta

from ..kafka import KafkaConnectionSettings, SchemaManagerSettings
from ._constants import (
    ADMIN_CLIENT_PREFIX,
    BROKER_PREFIX,
    EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL,
    EVENT_MANAGER_DEFAULT_KAFKA_TIMEOUT_MS,
)
from ._event_manager import (
    EventManager,
    KafkaEventManager,
    MockEventManager,
    NoopEventManager,
)

__all__ = [
    "BaseMetricsConfiguration",
    "DisabledMetricsConfiguration",
    "EventsConfiguration",
    "KafkaMetricsConfiguration",
    "MetricsConfiguration",
    "MockMetricsConfiguration",
    "metrics_configuration_factory",
]


class EventsConfiguration(BaseSettings):
    """Configuration for emitting events."""

    model_config = SettingsConfigDict(
        env_prefix="METRICS_", populate_by_name=True
    )

    topic_prefix: str = Field(
        "lsst.square.metrics.events",
        title="Metrics topic prefix",
        description=(
            "You probably should use the default here. It could be useful in"
            " development scenarios to change this."
        ),
        validation_alias=AliasChoices(
            "topicPrefix", "METRICS_EVENTS_TOPIC_PREFIX"
        ),
    )


def _require_bool(v: bool, wanted: bool) -> bool:  # noqa: FBT001
    """Pydantic validator to require a `bool` field have a particular value.

    Unfortunately, we cannot just use a `~typing.Literal` type because those
    effectively only work for strings. Conversion to other types is not done
    when the type is given as a literal, so we have to use this validator
    hack.
    """
    if v != wanted:
        raise ValueError(f"Input should be {wanted}")
    return v


class BaseMetricsConfiguration(BaseSettings, ABC):
    """Metrics configuration, including the required Kafka configuration.

    Currently, this only configures events, but if additional types of metrics
    are added in the future, that configuration will be added here.

    This is the recommended configuration approach if you don't need to use
    Kafka, the schema manager, or the schema registry in any other parts of
    your application. Applications that also use Kafka directly should instead
    create a `~safir.metrics.EventManager` with externally-managed Kafka
    clients.
    """

    model_config = SettingsConfigDict(populate_by_name=True)

    application: str = Field(
        ...,
        title="Application name",
        description=(
            "The name of the application that is emitting these metrics"
        ),
        validation_alias=AliasChoices("appName", "METRICS_APPLICATION"),
    )

    events: EventsConfiguration = Field(
        default_factory=EventsConfiguration,
        title="Events configuration",
    )

    @abstractmethod
    def make_manager(
        self,
        logger: BoundLogger | None = None,
        *,
        kafka_broker: KafkaBroker | None = None,
    ) -> EventManager:
        """Construct an `~safir.metrics.EventManager`.

        Parameters
        ----------
        logger
            Logger to use for internal logging. If not given, the
            ``safir.metrics`` logger will be used.
        kafka_broker
            Kafka broker to use. If not given, a new Kafka broker will be
            created and automatically closed when the event manager is closed.
            If a broker is provided, closing the event manager will have no
            effect on it and the caller is responsible for starting and
            closing it.

        Returns
        -------
        EventManager
            An event manager appropriate to the configuration.
        """


class DisabledMetricsConfiguration(BaseMetricsConfiguration):
    """Metrics configuration when metrics reporting is disabled."""

    model_config = SettingsConfigDict(extra="ignore", populate_by_name=True)

    enabled: Annotated[
        bool, AfterValidator(lambda x: _require_bool(x, False))
    ] = Field(
        ...,
        title="Whether to send events",
        description=(
            "If set to false, no events will be sent and all calls to publish"
            " events will be no-ops."
        ),
        validation_alias=AliasChoices("enabled", "METRICS_ENABLED"),
    )

    @override
    def make_manager(
        self,
        logger: BoundLogger | None = None,
        *,
        kafka_broker: KafkaBroker | None = None,
    ) -> NoopEventManager:
        if not logger:
            logger = structlog.get_logger("safir.metrics")
        return NoopEventManager(
            self.application, self.events.topic_prefix, logger
        )


class MockMetricsConfiguration(BaseMetricsConfiguration):
    """Metrics configuration when metrics publishing is mocked."""

    model_config = SettingsConfigDict(extra="ignore", populate_by_name=True)

    enabled: Annotated[
        bool, AfterValidator(lambda x: _require_bool(x, False))
    ] = Field(
        ...,
        title="Whether to send events",
        description=(
            "If set to false, no events will be sent and all calls to publish"
            " events will be no-ops."
        ),
        validation_alias=AliasChoices("enabled", "METRICS_ENABLED"),
    )

    mock: Annotated[bool, AfterValidator(lambda x: _require_bool(x, True))] = (
        Field(
            title="Mock publishers",
            description=(
                "If set to true, all event publishers will be"
                " unittest.mock.MagicMock instances which will record all"
                " calls to their publish methods."
            ),
            validation_alias=AliasChoices("mock", "METRICS_MOCK"),
        )
    )

    @override
    def make_manager(
        self,
        logger: BoundLogger | None = None,
        *,
        kafka_broker: KafkaBroker | None = None,
    ) -> MockEventManager:
        if not logger:
            logger = structlog.get_logger("safir.metrics")
        return MockEventManager(
            self.application, self.events.topic_prefix, logger
        )


class KafkaMetricsConfiguration(BaseMetricsConfiguration):
    """Metrics configuration when enabled, including Kafka configuration."""

    model_config = SettingsConfigDict(
        alias_generator=to_camel,
        extra="forbid",
        populate_by_name=True,
    )

    enabled: Annotated[
        bool, AfterValidator(lambda x: _require_bool(x, True))
    ] = Field(
        True,
        title="Whether to send events",
        description=(
            "If set to false, no events will be sent and all calls to publish"
            " events will be no-ops."
        ),
        validation_alias=AliasChoices("enabled", "METRICS_ENABLED"),
    )

    kafka: KafkaConnectionSettings = Field(
        default_factory=KafkaConnectionSettings,
        title="Kafka connection settings",
    )

    schema_manager: SchemaManagerSettings = Field(
        default_factory=SchemaManagerSettings,
        title="Kafka schema manager settings",
    )

    raise_on_error: bool = Field(
        False,
        title="Raise on error",
        description=(
            "True if we should raise an exception whenever there is an error"
            " with the metrics system dependencies, like Kafka or the Schema"
            " Manager. False if we should just log an error instead. This"
            " should be False for most production apps so that issues with the"
            " metrics infrastructure don't bring down the app."
        ),
    )

    backoff_interval: HumanTimedelta = Field(
        EVENT_MANAGER_DEFAULT_BACKOFF_INTERVAL,
        title="Backoff interval",
        description=(
            "The amount of time to wait until further operations are attempted"
            " when the event manager is in an error state."
        ),
    )

    @override
    def make_manager(
        self,
        logger: BoundLogger | None = None,
        *,
        kafka_broker: KafkaBroker | None = None,
        request_timeout_ms: int = EVENT_MANAGER_DEFAULT_KAFKA_TIMEOUT_MS,
        extra_broker_settings: dict[str, Any] | None = None,
        extra_admin_client_settings: dict[str, Any] | None = None,
    ) -> KafkaEventManager:
        """Make a KafkaEventManager.

        The request timeout is set low by default so we don't block the
        instrumented app for too long if kafka is unavailable.
        """
        extra_broker_settings = extra_broker_settings or {}
        extra_admin_client_settings = extra_admin_client_settings or {}

        manage_kafka_broker = kafka_broker is None
        if not kafka_broker:
            kafka_broker = KafkaBroker(
                client_id=f"{BROKER_PREFIX}-{self.application}",
                request_timeout_ms=request_timeout_ms,
                **self.kafka.to_faststream_params(),
                **extra_broker_settings,
            )
        kafka_admin_client = AIOKafkaAdminClient(
            client_id=f"{ADMIN_CLIENT_PREFIX}-{self.application}",
            request_timeout_ms=request_timeout_ms,
            **self.kafka.to_aiokafka_params(),
            **extra_admin_client_settings,
        )
        schema_manager = self.schema_manager.make_manager(logger=logger)

        return KafkaEventManager(
            application=self.application,
            topic_prefix=self.events.topic_prefix,
            kafka_broker=kafka_broker,
            kafka_admin_client=kafka_admin_client,
            schema_manager=schema_manager,
            manage_kafka_broker=manage_kafka_broker,
            raise_on_error=self.raise_on_error,
            backoff_interval=self.backoff_interval,
            logger=logger,
        )


type MetricsConfiguration = (
    MockMetricsConfiguration
    | DisabledMetricsConfiguration
    | KafkaMetricsConfiguration
)
"""Type to use for metrics configuration in the application config.

This will resolve to one of the various valid types of metrics configuration,
all of which support a `~safir.metrics.BaseMetricsConfiguration.make_manager`
method to create an `~safir.metrics.EventManager`.
"""


def metrics_configuration_factory() -> MetricsConfiguration:
    """Choose an appropriate metrics configuration based on the environment.

    This function is intended for use as the argument to the
    ``default_factory`` parameter to `pydantic.Field` for the application
    metrics configuration. It selects an appropriate metrics configuration
    based on which configuration class can be instantiated from the available
    environment variables. This is not necessary if the application
    configuration comes from a source such as YAML that specifies settings for
    the metrics configuration, since in that case Pydantic will correctly
    instantiate the correct settings model.

    Returns
    -------
    BaseMetricsConfiguration
        An appropriate metrics configuration.

    Raises
    ------
    pydantic.ValidationError
        Raised if none of the possible configurations have their required
        variables set.

    Examples
    --------
    .. code-block:: python

       from pydantic_settings import BaseSettings
       from safir.metrics import (
           MetricsConfiguration,
           metrics_configuration_factory,
       )


       class Config(BaseSettings):
           metrics: MetricsConfiguration = Field(
               default_factory=metrics_configuration_factory,
               title="Metrics configuration",
           )


       config = Config()
    """
    # When there are more possible configuration models, this code should
    # first try to instantiate all of the ones that require specific
    # environment variable settings to enable, and then finally
    # unconditionally try to return the default.
    try:
        return MockMetricsConfiguration()
    except ValidationError:
        try:
            return DisabledMetricsConfiguration()
        except ValidationError:
            return KafkaMetricsConfiguration()



"""Return internal objects as XML responses."""

from __future__ import annotations

import jinja2
from fastapi import Request, Response
from fastapi.templating import Jinja2Templates

from safir.datetime import isodatetime

from ._models import JobError

__all__ = ["UWSTemplates"]

_environment = jinja2.Environment(
    loader=jinja2.PackageLoader("safir", "uws/templates"),
    undefined=jinja2.StrictUndefined,
    autoescape=True,
)
_environment.filters["isodatetime"] = isodatetime
_templates = Jinja2Templates(env=_environment)


class UWSTemplates:
    """Template responses for the UWS protocol.

    This also includes VOSI-Availability since it was convenient to provide.
    """

    def error(self, request: Request, error: JobError) -> Response:
        """Return the error of a job as an XML response."""
        return _templates.TemplateResponse(
            request,
            "error.xml",
            {"error": error},
            media_type="application/xml",
        )



"""Standardized metadata for Roundtable and Phalanx HTTP services."""

from ._models import Metadata
from ._retrieve import get_metadata, get_project_url

__all__ = [
    "Metadata",
    "get_metadata",
    "get_project_url",
]



"""Test handling of jobs that fail."""

from __future__ import annotations

import pytest
from httpx import AsyncClient
from vo_models.uws import JobSummary

from safir.arq.uws import WorkerFatalError, WorkerTransientError
from safir.datetime import isodatetime
from safir.testing.slack import MockSlackWebhook
from safir.testing.uws import MockUWSJobRunner, assert_job_summary_equal
from safir.uws._dependencies import UWSFactory
from safir.uws._exceptions import TaskError

from ..support.uws import SimpleParameters, SimpleXmlParameters

ERRORED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>ERROR</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Sarah</uws:parameter>
  </uws:parameters>
  <uws:errorSummary type="{}" hasDetail="{}">
    <uws:message>{}</uws:message>
  </uws:errorSummary>
</uws:job>
"""

JOB_ERROR_SUMMARY = """
<?xml version="1.0" encoding="UTF-8"?>
<VOTABLE version="1.4" xmlns="http://www.ivoa.net/xml/VOTable/v1.3">
  <INFO name="QUERY_STATUS" value="ERROR">
{}
  </INFO>
</VOTABLE>
"""


@pytest.mark.asyncio
async def test_temporary_error(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(test_token, SimpleParameters(name="Sarah"))

    # The pending job has no error.
    r = await client.get("/test/jobs/1/error")
    assert r.status_code == 404

    # Execute the job.
    r = await client.post("/test/jobs/1/phase", data={"PHASE": "RUN"})
    assert r.status_code == 303
    await runner.mark_in_progress(test_token, "1")
    exc = WorkerTransientError("Something failed")
    result = TaskError.from_worker_error(exc)
    job = await runner.mark_complete(test_token, "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        ERRORED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            isodatetime(job.destruction_time),
            "transient",
            "false",
            "ServiceUnavailable: Something failed",
        ),
    )

    # Retrieve the error separately.
    r = await client.get("/test/jobs/1/error")
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "ServiceUnavailable: Something failed"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []


@pytest.mark.asyncio
async def test_fatal_error(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(test_token, SimpleParameters(name="Sarah"))

    # Start the job.
    r = await client.post("/test/jobs/1/phase", data={"PHASE": "RUN"})
    assert r.status_code == 303
    await runner.mark_in_progress(test_token, "1")
    exc = WorkerFatalError("Whoops", "Some details")
    result = TaskError.from_worker_error(exc)
    job = await runner.mark_complete(test_token, "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        ERRORED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            isodatetime(job.destruction_time),
            "fatal",
            "true",
            "Error: Whoops",
        ),
    )

    # Retrieve the error separately.
    r = await client.get("/test/jobs/1/error")
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "Error: Whoops\n\nSome details"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []


@pytest.mark.asyncio
async def test_unknown_error(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
    mock_slack: MockSlackWebhook,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(test_token, SimpleParameters(name="Sarah"))

    # Start the job.
    r = await client.post("/test/jobs/1/phase", data={"PHASE": "RUN"})
    assert r.status_code == 303
    await runner.mark_in_progress(test_token, "1")
    result = ValueError("Unknown exception")
    job = await runner.mark_complete(test_token, "1", result)

    # Check the results.
    assert job.start_time
    assert job.end_time
    r = await client.get("/test/jobs/1")
    assert r.status_code == 200
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        ERRORED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            isodatetime(job.destruction_time),
            "fatal",
            "true",
            "Error: Unknown error executing task",
        ),
    )

    # Retrieve the error separately.
    r = await client.get("/test/jobs/1/error")
    assert r.status_code == 200
    assert r.text == JOB_ERROR_SUMMARY.strip().format(
        "Error: Unknown error executing task\n\nValueError: Unknown exception"
    )

    # For now, this shouldn't have resulted in Slack errors.
    assert mock_slack.messages == []



"""Fixtures for tests of the testing support infrastructure."""

from __future__ import annotations

from collections.abc import Iterator
from datetime import timedelta
from pathlib import Path

import pytest

from safir.testing.gcs import MockStorageClient, patch_google_storage


@pytest.fixture
def mock_gcs_file() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        path=Path(__file__).parent,
        expected_expiration=timedelta(hours=1),
        bucket_name="some-bucket",
    )


@pytest.fixture
def mock_gcs_minimal() -> Iterator[MockStorageClient]:
    yield from patch_google_storage()



"""Mock Slack server for testing Slack messaging."""

from __future__ import annotations

import json
from typing import Any

import respx
from httpx import Request, Response
from pydantic import SecretStr

__all__ = ["MockSlackWebhook", "mock_slack_webhook"]


class MockSlackWebhook:
    """Represents a Slack incoming webhook and remembers what was posted.

    Parameters
    ----------
    url
        URL that the mock has been configured to listen on.

    Attributes
    ----------
    messages
        Messages that have been posted to the webhook so far.
    url
        URL that the mock has been configured to listen on.
    """

    def __init__(self, url: str) -> None:
        self.url = url
        self.messages: list[dict[str, Any]] = []

    def post_webhook(self, request: Request) -> Response:
        """Post a Slack message.

        The provided message is stored in the messages attribute.

        Parameters
        ----------
        request
            Incoming request.

        Returns
        -------
        httpx.Response
            Always returns a 201 response.
        """
        self.messages.append(json.loads(request.content.decode()))
        return Response(201)


def mock_slack_webhook(
    hook_url: str | SecretStr, respx_mock: respx.Router
) -> MockSlackWebhook:
    """Set up a mocked Slack server.

    Parameters
    ----------
    hook_url
        URL for the Slack incoming webhook to mock.
    respx_mock
        The mock router.

    Returns
    -------
    MockSlackWebhook
        The mock Slack webhook API object.

    Examples
    --------
    To set up this mock, put a fixture in :file:`conftest.py` similar to the
    following:

    .. code-block:: python

       import pytest
       import respx
       from safir.testing.slack import MockSlackWebhook, mock_slack_webhook


       @pytest.fixture
       def mock_slack(
           config: Config, respx_mock: respx.Router
       ) -> MockWebhookSlack:
           return mock_slack_webhook(config.slack_webhook, respx_mock)

    This uses respx_ to mock the Slack webhook URL obtained from the
    configuration of the application under test. Use it in a test as follows:

    .. code-block:: python

       @pytest.mark.asyncio
       def test_something(
           client: AsyncClient, mock_slack: MockSlackWebhook
       ) -> None:
           # Do something with client that generates Slack messages.
           assert mock_slack.messages == [{...}, {...}]
    """
    if isinstance(hook_url, SecretStr):
        hook_url = hook_url.get_secret_value()
    mock = MockSlackWebhook(hook_url)
    respx_mock.post(hook_url).mock(side_effect=mock.post_webhook)
    return mock



"""Exception helpers for Sentry instrumentation."""

from typing import Any, Self

from httpx import HTTPError, HTTPStatusError
from sentry_sdk.types import Event

__all__ = [
    "SentryException",
    "SentryWebException",
]


class SentryException(Exception):
    """Enriches the Sentry context when paired with the ``enrich`` handler."""

    def __init__(self, message: str) -> None:
        # Do not call the parent Exception constructor here, because calling
        # it with a different number of arguments than the constructor
        # argument of derived exceptions breaks pickling. See the end of
        # https://github.com/python/cpython/issues/44791. This requires
        # implementing __str__ rather than relying on the default behavior.
        #
        # Arguably, this is a bug in the __reduce__ method of BaseException
        # and its interaction with constructors, but it appears to be hard to
        # fix. See https://github.com/python/cpython/issues/76877.
        self.message = message
        self.tags: dict[str, str] = {}
        self.contexts: dict[str, dict[str, Any]] = {}

    def __str__(self) -> str:
        return self.message

    def enrich(self, event: Event) -> Event:
        """Merge our tags and contexts into the event's."""
        event["tags"] = event.setdefault("tags", {})
        event["tags"].update(self.tags)
        event["contexts"] = event.get("contexts", {}) | self.contexts
        return event


class SentryWebException(SentryException):
    """Parent class of exceptions arising from HTTPX_ failures.

    Captures additional information from any HTTPX_ exception.  Intended to be
    subclassed.

    Parameters
    ----------
    message
        Exception string value, which is the default Slack message.
    method
        Method of request.
    url
        URL of the request.
    user
        Username on whose behalf the request is being made.
    status
        Status code of failure, if any.
    body
        Body of failure message, if any.
    """

    @classmethod
    def from_exception(cls, exc: HTTPError, user: str | None = None) -> Self:
        """Create an exception from an HTTPX_ exception.

        Parameters
        ----------
        exc
            Exception from HTTPX.
        user
            User on whose behalf the request is being made, if known.

        Returns
        -------
        SlackWebException
            Newly-constructed exception.
        """
        if isinstance(exc, HTTPStatusError):
            status = exc.response.status_code
            method = exc.request.method
            message = f"Status {status} from {method} {exc.request.url}"
            return cls(
                message,
                method=exc.request.method,
                url=str(exc.request.url),
                user=user,
                status=status,
                body=exc.response.text,
            )
        else:
            exc_name = type(exc).__name__
            message = f"{exc_name}: {exc!s}" if str(exc) else exc_name

            # All httpx.HTTPError exceptions have a slot for the request,
            # initialized to None and then sometimes added by child
            # constructors or during exception processing. The request
            # property is a property method that raises RuntimeError if
            # request has not been set, so we can't just check for None. Hence
            # this approach of attempting to use the request and falling back
            # on reporting less data if that raised any exception.
            try:
                return cls(
                    message,
                    method=exc.request.method,
                    url=str(exc.request.url),
                    user=user,
                )
            except Exception:
                return cls(message, user=user)

    def __init__(
        self,
        message: str,
        *,
        method: str | None = None,
        url: str | None = None,
        user: str | None = None,
        status: int | None = None,
        body: str | None = None,
    ) -> None:
        super().__init__(message)
        self.method = method
        self.url = url
        self.status = status
        self.body = body
        self.user = user

        if self.method:
            self.tags["httpx_request_method"] = self.method
        if self.user:
            self.tags["gafaelfaw_user"] = self.user
        if self.url:
            self.tags["httpx_request_url"] = self.url
        if self.status:
            self.tags["httpx_request_status"] = str(self.status)
        if self.body:
            self.contexts["httpx_request_info"] = {"body": self.body}

    def __str__(self) -> str:
        result = self.message
        if self.body:
            result += f"\nBody:\n{self.body}\n"
        return result



"""Tests for the Google Cloud Storage support infrastructure.

These are just basic sanity checks that the mocking is working correctly and
the basic calls work.
"""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
from pathlib import Path

import google.auth
import pytest
from google.auth import impersonated_credentials
from google.cloud import storage

from safir.testing.gcs import MockStorageClient


def test_mock(mock_gcs: MockStorageClient) -> None:
    client = storage.Client()
    bucket = client.bucket("some-bucket")
    blob = bucket.blob("something")
    credentials = google.auth.default()
    signing_credentials = impersonated_credentials.Credentials(
        source_credentials=credentials,
        target_principal="some-service-account",
        target_scopes="https://www.googleapis.com/auth/devstorage.read_only",
        lifetime=2,
    )
    signed_url = blob.generate_signed_url(
        version="v4",
        expiration=timedelta(hours=1),
        method="GET",
        response_type="application/fits",
        credentials=signing_credentials,
    )
    assert signed_url == "https://example.com/something"

    # The wrong expiration produces an assertion.
    with pytest.raises(AssertionError):
        blob.generate_signed_url(
            version="v4",
            expiration=timedelta(hours=2),
            method="GET",
            response_type="application/fits",
            credentials=signing_credentials,
        )

    # The wrong bucket produces an assertion.
    with pytest.raises(AssertionError):
        bucket = client.bucket("wrong-bucket")


def test_mock_files(mock_gcs_file: MockStorageClient) -> None:
    this_file = Path(__file__)
    client = storage.Client()
    bucket = client.bucket("some-bucket")
    blob = bucket.blob(this_file.name)

    # Test that signed URLs still work.
    credentials = google.auth.default()
    signing_credentials = impersonated_credentials.Credentials(
        source_credentials=credentials,
        target_principal="some-service-account",
        target_scopes="https://www.googleapis.com/auth/devstorage.read_only",
        lifetime=2,
    )
    signed_url = blob.generate_signed_url(
        version="v4",
        expiration=timedelta(hours=1),
        method="GET",
        response_type="application/fits",
        credentials=signing_credentials,
    )
    assert signed_url == f"https://example.com/{this_file.name}"

    # Test the file-specific methods.
    assert blob.exists()
    assert blob.size == this_file.stat().st_size
    assert blob.updated == datetime.fromtimestamp(
        this_file.stat().st_mtime, tz=UTC
    )
    assert blob.etag == str(this_file.stat().st_ino)
    assert blob.download_as_bytes() == this_file.read_bytes()
    with blob.open("rb") as f:
        contents = f.read()
    assert contents == this_file.read_bytes()

    # Test an invalid open mode.
    with pytest.raises(AssertionError):
        blob.open("wb")

    # Test a nonexistent file.
    blob = bucket.blob("does-not-exist")
    assert not blob.exists()


def test_mock_minimal(mock_gcs_minimal: MockStorageClient) -> None:
    """Minimal configuration, which doesn't check lifetime or bucket name."""
    client = storage.Client()
    bucket = client.bucket("some-bucket")

    # It doesn't matter what bucket name we choose, since we didn't request
    # verification.
    bucket = client.bucket("other-bucket")

    # It doesn't matter what expiration we specify on signed URLs.
    blob = bucket.blob("a-file")
    signed_url = blob.generate_signed_url(
        version="v4", expiration=timedelta(minutes=1), method="GET"
    )
    assert signed_url == "https://example.com/a-file"
    signed_url = blob.generate_signed_url(
        version="v4", expiration=timedelta(hours=1), method="GET"
    )
    assert signed_url == "https://example.com/a-file"



"""Configuration for the UWS service."""

from __future__ import annotations

from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from datetime import datetime, timedelta

from arq.connections import RedisSettings
from pydantic import Field, HttpUrl, SecretStr
from pydantic_settings import BaseSettings
from vo_models.uws import JobSummary

from safir.arq import ArqMode, build_arq_redis_settings
from safir.pydantic import EnvRedisDsn, HumanTimedelta, SecondsTimedelta

from ._models import Job, ParametersModel

type DestructionValidator = Callable[[datetime, Job], datetime]
"""Type for a validator for a new destruction time."""

type ExecutionDurationValidator = Callable[
    [timedelta | None, Job], timedelta | None
]
"""Type for a validator for a new execution duration."""

__all__ = [
    "UWSAppSettings",
    "UWSConfig",
    "UWSRoute",
]


@dataclass
class UWSRoute:
    """Defines a FastAPI dependency to get the UWS job parameters."""

    dependency: Callable[..., Coroutine[None, None, ParametersModel]]
    """Type for a dependency that gathers parameters for a job."""

    summary: str
    """Summary string for API documentation."""

    description: str | None = None
    """Description string for API documentation."""


@dataclass
class UWSConfig:
    """Configuration for the UWS service.

    The UWS service may be embedded in a variety of VO applications. This
    class encapsulates the configuration of the UWS component that may vary by
    service or specific installation.
    """

    arq_mode: ArqMode
    """What mode to use for the arq queue."""

    arq_redis_settings: RedisSettings
    """Settings for Redis for the arq queue."""

    async_post_route: UWSRoute
    """Route configuration for creating an async job via POST.

    The FastAPI dependency included in this object should expect POST
    parameters and return the Pydantic model of the job parameters.
    """

    execution_duration: timedelta
    """Maximum execution time in seconds.

    Jobs that run longer than this length of time will be automatically
    aborted.
    """

    job_summary_type: type[JobSummary]
    """Type representing the parameter-qualified job summary type.

    Must be set to `~vo_models.uws.models.JobSummary` qualified with the
    appropriate subclass of `~vo_models.uws.models.Parameters`. This is
    necessary to work around limitations in pydantic-xml, which require the
    types to be known at class instantiation time.
    """

    lifetime: timedelta
    """The lifetime of jobs.

    After this much time elapses since the creation of the job, all of the
    results from the job will be cleaned up and all record of the job will be
    deleted.
    """

    parameters_type: type[ParametersModel]
    """Type representing the job parameters.

    This will be used to validate parameters and to parse them before passing
    them to the worker.
    """

    signing_service_account: str
    """Email of service account to use for signed URLs.

    The default credentials that the application frontend runs with must have
    the ``roles/iam.serviceAccountTokenCreator`` role on the service account
    with this email.
    """

    wobbly_url: str | HttpUrl
    """URL to the Wobbly UWS job tracking API."""

    worker: str
    """Name of the backend worker to call to execute a job."""

    slack_webhook: SecretStr | None = None
    """Slack incoming webhook for reporting errors."""

    sync_get_route: UWSRoute | None = None
    """Route configuration for creating a sync job via GET.

    The FastAPI dependency included in this object should expect GET
    parameters and the Pydantic model of the job parameters. If `None`, no
    route to create a job via sync GET will be created.
    """

    sync_post_route: UWSRoute | None = None
    """Route configuration for creating a sync job via POST.

    The FastAPI dependency included in this object should expect POST
    parameters and return the Pydantic model of the job parameters. If `None`,
    no route to create a job via sync POST will be created.
    """

    sync_timeout: timedelta = timedelta(minutes=5)
    """Maximum lifetime of a sync request."""

    url_lifetime: timedelta = timedelta(minutes=15)
    """How long result URLs should be valid for."""

    validate_destruction: DestructionValidator | None = None
    """Validate a new destruction time for a job.

    If provided, called with the requested destruction time and the current
    job record and should return the new destruction time. Otherwise, any
    destruction time before the configured maximum lifetime will be allowed.
    """

    validate_execution_duration: ExecutionDurationValidator | None = None
    """Validate a new execution duration for a job.

    If provided, called with the requested execution duration and the current
    job record and should return the new execution duration time. Otherwise,
    the execution duration may not be changed.
    """

    wait_timeout: timedelta = timedelta(minutes=1)
    """Maximum time a client can wait for a job change."""


class UWSAppSettings(BaseSettings):
    """Settings common to all applications using the UWS library.

    The ``Config`` class for an application should inherit from this class to
    get the standard UWS application settings.
    """

    arq_mode: ArqMode = Field(
        ArqMode.production,
        title="arq operation mode",
        description="This will always be production outside the test suite",
    )

    arq_queue_url: EnvRedisDsn = Field(
        ...,
        title="arq Redis DSN",
        description="DSN of Redis server to use for the arq queue",
    )

    arq_queue_password: SecretStr | None = Field(
        None,
        title="Password for arq Redis server",
        description="Password of Redis server to use for the arq queue",
    )

    grace_period: SecondsTimedelta = Field(
        timedelta(seconds=30),
        title="Grace period for jobs",
        description=(
            "How long to wait for a job to finish on shutdown before"
            " canceling it"
        ),
    )

    lifetime: HumanTimedelta = Field(
        timedelta(days=7), title="Lifetime of job results"
    )

    service_account: str = Field(
        ...,
        title="Service account for URL signing",
        description=(
            "Email of the service account to use for signed URLs of results."
            " The default credentials that the application frontend runs with"
            " must have the ``roles/iam.serviceAccountTokenCreator`` role on"
            " the service account with this email."
        ),
    )

    storage_url: str = Field(
        ...,
        title="Root URL for cutout results",
        description=(
            "Must be a ``gs`` or ``s3`` URL pointing to a Google Cloud Storage"
            " bucket that is writable by the backend and readable by the"
            " frontend."
        ),
    )

    sync_timeout: HumanTimedelta = Field(
        timedelta(minutes=1), title="Timeout for sync requests"
    )

    timeout: SecondsTimedelta = Field(
        timedelta(minutes=10),
        title="Job timeout in seconds",
        description=(
            "Must be given as a number of seconds as a string or integer"
        ),
    )

    wobbly_url: HttpUrl = Field(
        ...,
        title="Wobbly URL",
        description="URL to Wobbly UWS job tracking API",
    )

    @property
    def arq_redis_settings(self) -> RedisSettings:
        """Redis settings for arq."""
        return build_arq_redis_settings(
            self.arq_queue_url, self.arq_queue_password
        )

    def build_uws_config(
        self,
        *,
        async_post_route: UWSRoute,
        job_summary_type: type[JobSummary],
        parameters_type: type[ParametersModel],
        slack_webhook: SecretStr | None = None,
        sync_get_route: UWSRoute | None = None,
        sync_post_route: UWSRoute | None = None,
        url_lifetime: timedelta = timedelta(minutes=15),
        validate_destruction: DestructionValidator | None = None,
        validate_execution_duration: ExecutionDurationValidator | None = None,
        wait_timeout: timedelta = timedelta(minutes=1),
        worker: str,
    ) -> UWSConfig:
        """Construct a `UWSConfig` object from the application configuration.

        This helper method can be used by application ``Config`` classes to
        help build the `UWSConfig` object corresponding to the application
        configuration. Its parameters are the additional settings accepted by
        the UWS library that are not part of the ``UWSAppSettings`` model.

        Parameters
        ----------
        async_post_route
            Route configuration for job parameters for an async job via
            POST. The FastAPI dependency included in this object should expect
            POST parameters and return a Pydantic model representing the job
            parameters.
        job_summary_type
            Type representing the XML job summary type, qualified with an
            appropriate subclass of `~vo_models.uws.models.Parameters`. That
            subclass should be the same as that returned by the
            ``to_xml_model`` method of ``parameters_type``.
        parameters_type
            Type representing the job parameters. This will be used to
            validate parameters and to parse them before passing them to the
            worker.
        slack_webhook
            Slack incoming webhook for reporting errors.
        sync_get_route
            Route configuration for creating a sync job via GET. The FastAPI
            dependency included in this object should expect GET parameters
            and return a Pydantic model representing the job parameters. If
            `None`, no route to create a job via sync GET will be created.
        sync_post_route
            Route configuration for creating a sync job via POST. The FastAPI
            dependency included in this object should expect POST parameters
            and return a Pydantic model representing the job parameters. If
            `None`, no route to create a job via sync POST will be created.
        url_lifetime
            How long result URLs should be valid for.
        validate_destruction
            Validate a new destruction time for a job. If provided, called
            with the requested destruction time and the current job record and
            should return the new destruction time. Otherwise, any destruction
            time before the configured maximum lifetime will be allowed.
        validate_execution_duration
            Validate a new execution duration for a job. If provided, called
            with the requested execution duration and the current job record
            and should return the new execution duration time. Otherwise, the
            execution duration may not be changed.
        wait_timeout
            Maximum time a client can wait for a job change.
        worker
            Name of the backend worker to call to execute a job.

        Returns
        -------
        UWSConfig
            UWS configuration.

        Examples
        --------
        Normally, this method is used from a property method that returns the
        UWS configuration, such as the following example for a cutout service:

        .. code-block:: python

           @property
           def uws_config(self) -> UWSConfig:
               return self.build_uws_config(
                   async_post_route=UWSRoute(
                       dependency=post_params_dependency,
                       summary="Create async cutout job",
                   ),
                   parameters_type=CutoutParameters,
                   slack_webhook=self.slack_webhook,
                   sync_get_route=UWSRoute(
                       dependency=get_params_dependency,
                       summary="Synchronous cutout",
                   ),
                   sync_post_route=UWSRoute(
                       dependency=post_params_dependency,
                       summary="Synchronous cutout",
                   ),
                   worker="cutout",
               )
        """
        return UWSConfig(
            arq_mode=self.arq_mode,
            arq_redis_settings=self.arq_redis_settings,
            execution_duration=self.timeout,
            job_summary_type=job_summary_type,
            lifetime=self.lifetime,
            parameters_type=parameters_type,
            signing_service_account=self.service_account,
            worker=worker,
            slack_webhook=slack_webhook,
            sync_timeout=self.sync_timeout,
            async_post_route=async_post_route,
            sync_get_route=sync_get_route,
            sync_post_route=sync_post_route,
            url_lifetime=url_lifetime,
            validate_destruction=validate_destruction,
            validate_execution_duration=validate_execution_duration,
            wait_timeout=wait_timeout,
            wobbly_url=self.wobbly_url,
        )



"""Tests for the safir.redis module."""

from __future__ import annotations

import pytest
import redis.asyncio as redis
from cryptography.fernet import Fernet
from pydantic import BaseModel, Field, SecretStr

from safir.redis import (
    DeserializeError,
    EncryptedPydanticRedisStorage,
    PydanticRedisStorage,
)
from safir.slack.blockkit import SlackTextField


class DemoModel(BaseModel):
    """A demo model for testing."""

    name: str | None = Field(..., description="The name of the model.")
    value: int = Field(4, description="The value of the model.")


async def basic_testing(storage: PydanticRedisStorage[DemoModel]) -> None:
    """Test basic storage operations for either encrypted or unencrypted
    storage.
    """
    await storage.store("mark42", DemoModel(name="Mark", value=42))
    await storage.store("mark13", DemoModel(name="Mark", value=13))
    await storage.store("jon7", DemoModel(name="Jon", value=7))

    assert await storage.get("mark42") == DemoModel(name="Mark", value=42)
    async for key in storage.scan("mark*"):
        assert key in ["mark13", "mark42"]

    assert await storage.delete("mark42") is True
    assert [m async for m in storage.scan("mark*")] == ["mark13"]
    assert await storage.get("mark42") is None
    assert await storage.delete("mark42") is False

    await storage.delete_all("mark*")
    assert [m async for m in storage.scan("mark*")] == []

    await storage.delete_all("*")
    assert await storage.get("jon7") is None


@pytest.mark.asyncio
async def test_pydantic_redis_storage(redis_client: redis.Redis) -> None:
    """Test unencrypted storage with the DemoModel."""
    storage = PydanticRedisStorage(datatype=DemoModel, redis=redis_client)
    await basic_testing(storage)


@pytest.mark.asyncio
async def test_pydantic_redis_storage_with_prefix(
    redis_client: redis.Redis,
) -> None:
    """Test unencrypted storage with the DemoModel."""
    storage = PydanticRedisStorage(
        datatype=DemoModel, redis=redis_client, key_prefix="test:"
    )
    await basic_testing(storage)


@pytest.mark.asyncio
async def test_encrypted_pydantic_redis_storage(
    redis_client: redis.Redis,
) -> None:
    """Test encrypted storage."""
    storage = EncryptedPydanticRedisStorage(
        datatype=DemoModel,
        redis=redis_client,
        encryption_key=Fernet.generate_key().decode(),
    )
    await basic_testing(storage)


@pytest.mark.asyncio
async def test_pydantic_exclude(redis_client: redis.Redis) -> None:
    """Test exclude parameters to store."""
    storage = PydanticRedisStorage(datatype=DemoModel, redis=redis_client)
    none_explicit = DemoModel(name=None, value=4)
    none_no_value = DemoModel(name=None)

    await storage.store("test", none_explicit)
    assert await redis_client.get("test") == b'{"name":null,"value":4}'
    await storage.store("test", none_no_value)
    assert await redis_client.get("test") == b'{"name":null,"value":4}'

    await storage.store("test", none_explicit, exclude_unset=True)
    assert await redis_client.get("test") == b'{"name":null,"value":4}'
    await storage.store("test", none_no_value, exclude_unset=True)
    assert await redis_client.get("test") == b'{"name":null}'

    await storage.store("test", none_explicit, exclude_defaults=True)
    assert await redis_client.get("test") == b'{"name":null}'
    await storage.store("test", none_no_value, exclude_defaults=True)
    assert await redis_client.get("test") == b'{"name":null}'

    await storage.store("test", none_explicit, exclude_none=True)
    assert await redis_client.get("test") == b'{"value":4}'
    await storage.store("test", none_no_value, exclude_none=True)
    assert await redis_client.get("test") == b'{"value":4}'

    await storage.store(
        "test", none_explicit, exclude_none=True, exclude_defaults=True
    )
    assert await redis_client.get("test") == b"{}"

    await storage.store("mark42", DemoModel(name="Mark", value=42))
    await storage.store("mark13", DemoModel(name="Mark", value=13))


@pytest.mark.asyncio
async def test_encrypted_pydantic_redis_storage_secretstr(
    redis_client: redis.Redis,
) -> None:
    """Test that the secret can be passed as str or SecretStr."""
    encryption_key = Fernet.generate_key().decode()
    storage = EncryptedPydanticRedisStorage(
        datatype=DemoModel, redis=redis_client, encryption_key=encryption_key
    )

    await storage.store("mark42", DemoModel(name="Mark", value=42))

    storage = EncryptedPydanticRedisStorage(
        datatype=DemoModel,
        redis=redis_client,
        encryption_key=SecretStr(encryption_key),
    )

    assert await storage.get("mark42") == DemoModel(name="Mark", value=42)


class PetModel(BaseModel):
    id: int
    name: str
    age: int


class CustomerModel(BaseModel):
    id: int
    name: str
    email: str


@pytest.mark.asyncio
async def test_multiple_stores(redis_client: redis.Redis) -> None:
    """Test multiple stores with unique prefixes for each."""
    pet_store = PydanticRedisStorage(
        datatype=PetModel,
        redis=redis_client,
        key_prefix="pet:",
    )
    customer_store = PydanticRedisStorage(
        datatype=CustomerModel,
        redis=redis_client,
        key_prefix="customer:",
    )

    await pet_store.store("emma", PetModel(id=1, name="Emma", age=2))
    await customer_store.store(
        "emma", CustomerModel(id=1, name="Emma", email="emma@example.com")
    )

    assert await pet_store.get("emma") == PetModel(id=1, name="Emma", age=2)
    assert await customer_store.get("emma") == CustomerModel(
        id=1, name="Emma", email="emma@example.com"
    )

    # Scanned keys should not contain the prefix.
    assert [m async for m in pet_store.scan("*")] == ["emma"]
    assert [m async for m in customer_store.scan("*")] == ["emma"]

    await pet_store.delete_all("*")
    # Pet emma should be gone, but customer emma should still be there.
    assert await pet_store.delete("emma") is False
    assert await customer_store.delete("emma") is True


@pytest.mark.asyncio
async def test_deserialization_error(redis_client: redis.Redis) -> None:
    """Test that deserialization errors are caught."""
    storage = PydanticRedisStorage(datatype=DemoModel, redis=redis_client)
    await storage._redis.set("key", b"not a valid model")
    with pytest.raises(
        DeserializeError, match="^Cannot deserialize data for key key"
    ) as e:
        await storage.get("key")
    slack_message = e.value.to_slack()
    assert slack_message.fields[-1] == SlackTextField(
        heading="Key", text="key"
    )
    await storage.delete_all("*")


@pytest.mark.asyncio
async def test_deserialization_error_with_key_prefix(
    redis_client: redis.Redis,
) -> None:
    """Test that deserialization errors are presented correctly when a key
    prefix is used.
    """
    storage = PydanticRedisStorage(
        datatype=DemoModel, redis=redis_client, key_prefix="test:"
    )
    await storage._redis.set("test:key", b"not a valid model")
    with pytest.raises(
        DeserializeError, match="^Cannot deserialize data for key test:key"
    ) as e:
        await storage.get("key")
    slack_message = e.value.to_slack()
    assert slack_message.fields[-1] == SlackTextField(
        heading="Key", text="test:key"
    )
    await storage.delete_all("*")



"""Functions to get the current date and time."""

from __future__ import annotations

from datetime import UTC, datetime

__all__ = ["current_datetime"]


def current_datetime(*, microseconds: bool = False) -> datetime:
    """Construct a `~datetime.datetime` for the current time.

    It's easy to forget to force all `~datetime.datetime` objects to be time
    zone aware. This function forces UTC for all objects.

    Databases do not always store microseconds in time fields, and having some
    dates with microseconds and others without them can lead to bugs, so by
    default it suppresses the microseconds.

    Parameters
    ----------
    microseconds
        Whether to include microseconds. Consider setting this to `True` when
        getting timestamps for error reporting, since granular timestamps can
        help in understanding sequencing.

    Returns
    -------
    datetime.datetime
        The current time forced to UTC and optionally with the microseconds
        field zeroed.
    """
    result = datetime.now(tz=UTC)
    if microseconds:
        return result
    else:
        return result.replace(microsecond=0)



"""Functions to simplify Kubernetes objects for comparison."""

from __future__ import annotations

from typing import Any

__all__ = ["strip_none"]


def strip_none(model: dict[str, Any]) -> dict[str, Any]:
    """Strip `None` values from a serialized Kubernetes object.

    Comparing Kubernetes objects against serialized expected output is a bit
    of a pain, since Kubernetes objects often contain tons of optional
    parameters and the ``to_dict`` serialization includes every parameter.
    The naive result is therefore tedious to read or understand.

    This function works around this by taking a serialized Kubernetes object
    and dropping all of the parameters that are set to `None`. The ``to_dict``
    form of a Kubernetes object should be passed through it first before
    comparing to the expected output.

    Parmaters
    ---------
    model
        Kubernetes model serialized with ``to_dict``.

    Returns
    -------
    dict
        Cleaned-up model with `None` parameters removed.
    """
    result = {}
    for key, value in model.items():
        if value is None:
            continue
        new_value = value
        if isinstance(value, dict):
            new_value = strip_none(value)
        elif isinstance(value, list):
            list_result = []
            for item in value:
                if isinstance(item, dict):
                    list_result.append(strip_none(item))
                else:
                    list_result.append(item)
            new_value = list_result
        result[key] = new_value
    return result



"""Storage layer for the UWS implementation."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import Any

from httpx import AsyncClient, HTTPError, Response
from pydantic import BaseModel
from vo_models.uws.types import ErrorType, ExecutionPhase

from safir.arq import JobMetadata
from safir.arq import JobResult as ArqJobResult
from safir.datetime import current_datetime, isodatetime

from ._config import UWSConfig
from ._exceptions import TaskError, UnknownJobError, WobblyError
from ._models import (
    Job,
    JobCreate,
    JobError,
    JobResult,
    JobUpdateAborted,
    JobUpdateCompleted,
    JobUpdateError,
    JobUpdateExecuting,
    JobUpdateMetadata,
    JobUpdateQueued,
    ParametersModel,
    SerializedJob,
)

__all__ = ["JobStore"]


class JobStore:
    """Stores and manipulates jobs in the database.

    The canonical representation of any UWS job is in the database. This class
    provides methods to create, update, and delete UWS job records and their
    associated results and errors.

    Parameters
    ----------
    config
        UWS configuration.
    http_client
        HTTP client to use to talk to Wobbly.
    """

    def __init__(self, config: UWSConfig, http_client: AsyncClient) -> None:
        self._config = config
        self._client = http_client
        self._base_url = str(config.wobbly_url).rstrip("/")

    async def create(
        self,
        token: str,
        *,
        run_id: str | None = None,
        parameters: ParametersModel,
        execution_duration: timedelta,
        lifetime: timedelta,
    ) -> Job:
        """Create a record of a new job.

        The job will be created in pending status.

        Parameters
        ----------
        token
            Token for an individual user.
        run_id
            A client-supplied opaque identifier to record with the job.
        parameters
            The input parameters to the job.
        execution_duration
            The maximum length of time for which a job is allowed to run in
            seconds.
        lifetime
            The maximum lifetime of the job and its results, in seconds.
            After this time, any record of the job will be deleted.

        Returns
        -------
        Job
            Newly-created job.

        Raises
        ------
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        job_create = JobCreate(
            json_parameters=parameters.model_dump(mode="json"),
            run_id=run_id,
            destruction_time=current_datetime() + lifetime,
            execution_duration=execution_duration,
        )
        r = await self._request("POST", token, body=job_create)
        job = SerializedJob.model_validate(r.json())
        return Job.from_serialized_job(job, self._config.parameters_type)

    async def delete(self, token: str, job_id: str) -> None:
        """Delete a job by ID.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Job ID to delete.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        await self._request("DELETE", token, job_id)

    async def get(self, token: str, job_id: str) -> Job:
        """Retrieve a job by ID.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Job ID to retrieve.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        r = await self._request("GET", token, job_id)
        job = SerializedJob.model_validate(r.json())
        return Job.from_serialized_job(job, self._config.parameters_type)

    async def list_jobs(
        self,
        token: str,
        *,
        phases: Iterable[ExecutionPhase] | None = None,
        after: datetime | None = None,
        count: int | None = None,
    ) -> list[SerializedJob]:
        """List the jobs for a particular user.

        Parameters
        ----------
        token
            Token for an individual user.
        phases
            Limit the result to jobs in this list of possible execution
            phases.
        after
            Limit the result to jobs created after the given datetime in UTC.
        count
            Limit the results to the most recent count jobs.

        Returns
        -------
        list of SerializedJob
            List of job descriptions matching the search criteria.

        Raises
        ------
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        query: list[tuple[str, str]] = []
        if phases:
            query.extend(("phase", p.value) for p in phases)
        if after:
            query.append(("since", isodatetime(after)))
        if count:
            query.append(("limit", str(count)))
        r = await self._request("GET", token, query=query)
        return [SerializedJob.model_validate(j) for j in r.json()]

    async def mark_aborted(self, token: str, job_id: str) -> None:
        """Mark a job as aborted.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        update = JobUpdateAborted(phase=ExecutionPhase.ABORTED)
        await self._request("PATCH", token, job_id, body=update)

    async def mark_completed(
        self, token: str, job_id: str, job_result: ArqJobResult
    ) -> None:
        """Mark a job as completed.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.
        job_result
            Result of the job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        if isinstance(job_result.result, Exception):
            await self.mark_failed(token, job_id, job_result.result)
            return
        results = [JobResult.from_worker_result(r) for r in job_result.result]
        update = JobUpdateCompleted(
            phase=ExecutionPhase.COMPLETED, results=results
        )
        await self._request("PATCH", token, job_id, body=update)

    async def mark_failed(
        self, token: str, job_id: str, exc: Exception
    ) -> None:
        """Mark a job as failed with an error.

        Currently, only one error is supported, even though Wobbly supports
        associating multiple errors with a job.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.
        exc
            Exception of failed job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        if isinstance(exc, TaskError):
            error = exc.to_job_error()
        else:
            error = JobError(
                type=ErrorType.FATAL,
                code="Error",
                message="Unknown error executing task",
                detail=f"{type(exc).__name__}: {exc!s}",
            )
        update = JobUpdateError(phase=ExecutionPhase.ERROR, errors=[error])
        await self._request("PATCH", token, job_id, body=update)

    async def mark_executing(
        self, token: str, job_id: str, start_time: datetime
    ) -> None:
        """Mark a job as executing.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.
        start_time
            Time at which the job started executing.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        update = JobUpdateExecuting(
            phase=ExecutionPhase.EXECUTING, start_time=start_time
        )
        await self._request("PATCH", token, job_id, body=update)

    async def mark_queued(
        self, token: str, job_id: str, metadata: JobMetadata
    ) -> None:
        """Mark a job as queued for processing.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.
        metadata
            Metadata about the underlying arq job.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        update = JobUpdateQueued(
            phase=ExecutionPhase.QUEUED, message_id=metadata.id
        )
        await self._request("PATCH", token, job_id, body=update)

    async def update_metadata(
        self,
        token: str,
        job_id: str,
        metadata: JobUpdateMetadata,
    ) -> None:
        """Update the destruction time or execution duration of a job.

        Parameters
        ----------
        token
            Token for an individual user.
        job_id
            Identifier of the job.
        metadata
            New job metadata.

        Raises
        ------
        UnknownJobError
            Raised if the job was not found.
        WobblyError
            Raised if the Wobbly request fails or returns a failure status.
        """
        await self._request("PATCH", token, job_id, body=metadata)

    async def _request(
        self,
        method: str,
        token: str,
        job_id: str | None = None,
        *,
        body: BaseModel | None = None,
        query: list[tuple[str, str]] | None = None,
    ) -> Response:
        """Send an HTTP request to Wobbly.

        Parameters
        ----------
        method
            HTTP method.
        token
            Token for an individual user.
        job_id
            Identifier of job to act on.
        body
            If given, a Pydantic model that should be serialized bo create the
            JSON body of the request.
        query
            If given, query parameters to send.

        Returns
        -------
        Response
            HTTP response object.

        Raises
        ------
        UnknownJobError
            Raised if the request was for a specific job and that job was not
            found.
        WobblyError
            Raised if the HTTP request fails or returns a failure status.
        """
        kwargs: dict[str, Any] = {
            "headers": {"Authorization": f"bearer {token}"}
        }
        if body:
            kwargs["json"] = body.model_dump(mode="json")
        if query:
            kwargs["params"] = query
        url = self._base_url + "/jobs"
        if job_id:
            url += "/" + job_id
        try:
            r = await self._client.request(method, url, **kwargs)
            if r.status_code == 404 and job_id:
                raise UnknownJobError(job_id)
            r.raise_for_status()
        except HTTPError as e:
            raise WobblyError.from_exception(e) from e
        return r



"""Sentry helpers."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta

from sentry_sdk.tracing import Span
from sentry_sdk.types import Event, Hint

from ._exceptions import SentryException

__all__ = [
    "before_send_handler",
    "duration",
    "fingerprint_env_handler",
    "sentry_exception_handler",
]


def duration(span: Span) -> timedelta:
    """Return the time spent in a span (to the present if not finished)."""
    if span.timestamp is None:
        timestamp = datetime.now(tz=UTC)
    else:
        timestamp = span.timestamp
    return timestamp - span.start_timestamp


def fingerprint_env_handler(event: Event, _: Hint) -> Event:
    """Add the environment to the event fingerprint.

    Without doing this, Sentry groups events from all environments into the
    same issue, and alerts that notify on new issues won't notify on a prod
    event if an issue has already been created from an event from another env
    :(

    https://github.com/getsentry/sentry/issues/64354
    """
    env = event.get("environment")
    if env is None:
        env = "no environment"
    fingerprint = event.get("fingerprint", [])
    event["fingerprint"] = [
        "{{ default }}",
        *fingerprint,
        env,
    ]
    return event


def sentry_exception_handler(event: Event, hint: Hint) -> Event:
    """Add tags and context from `~safir.sentry.SentryException`."""
    if exc_info := hint.get("exc_info"):
        exc = exc_info[1]
        if isinstance(exc, SentryException):
            exc.enrich(event)
    return event


def before_send_handler(event: Event, hint: Hint) -> Event:
    """Add the env to the fingerprint, and enrich from
    `~safir.sentry.SentryException`.
    """
    event = fingerprint_env_handler(event, hint)
    return sentry_exception_handler(event, hint)



"""Error models for FastAPI applications."""

from enum import StrEnum

from pydantic import BaseModel, Field

__all__ = [
    "ErrorDetail",
    "ErrorLocation",
    "ErrorModel",
]


class ErrorLocation(StrEnum):
    """Possible locations for an error.

    The first element of ``loc`` in `ErrorDetail` should be chosen from one of
    these values.
    """

    body = "body"
    header = "header"
    path = "path"
    query = "query"


class ErrorDetail(BaseModel):
    """The detail of the error message."""

    loc: list[str] | None = Field(
        None, title="Location", examples=[["area", "field"]]
    )

    msg: str = Field(..., title="Message", examples=["Some error messge"])

    type: str = Field(..., title="Error type", examples=["some_code"])


class ErrorModel(BaseModel):
    """A structured API error message."""

    detail: list[ErrorDetail] = Field(..., title="Detail")



"""Tests for Click utility functions."""

from __future__ import annotations

import click
from click.testing import CliRunner

from safir.click import display_help


def test_display_help() -> None:
    @click.group()
    def main() -> None:
        """Run some command."""

    @main.command()
    @click.argument("topic", default=None, required=False, nargs=1)
    @click.argument("subtopic", default=None, required=False, nargs=1)
    @click.pass_context
    def help(
        ctx: click.Context, topic: str | None, subtopic: str | None
    ) -> None:
        """Show help for any command."""
        display_help(main, ctx, topic, subtopic)

    @main.command()
    def foo() -> None:
        """Run foo."""
        click.echo("Ran foo")

    @main.group()
    def bar() -> None:
        """Bar commands."""

    @bar.command()
    def something() -> None:
        """Do something with a bar."""

    runner = CliRunner()
    result = runner.invoke(main, ["help"], catch_exceptions=False)
    assert "Run some command" in result.output
    result = runner.invoke(main, ["help", "foo"], catch_exceptions=False)
    assert "main foo [OPTIONS]" in result.output
    assert "Run foo" in result.output
    result = runner.invoke(
        main, ["help", "bar", "something"], catch_exceptions=False
    )
    assert "main bar something [OPTIONS]" in result.output
    assert "Do something with a bar" in result.output



"""Database transaction retry support."""

import asyncio
import contextlib
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import overload

from sqlalchemy.exc import DBAPIError

__all__ = ["retry_async_transaction"]


@overload
def retry_async_transaction[**P, T](
    __func: Callable[P, Coroutine[None, None, T]], /
) -> Callable[P, Coroutine[None, None, T]]: ...


@overload
def retry_async_transaction[**P, T](
    *, delay: float = 0.5, max_tries: int = 3
) -> Callable[
    [Callable[P, Coroutine[None, None, T]]],
    Callable[P, Coroutine[None, None, T]],
]: ...


def retry_async_transaction[**P, T](
    __func: Callable[P, Coroutine[None, None, T]] | None = None,
    /,
    *,
    delay: float = 0.5,
    max_tries: int = 3,
) -> (
    Callable[P, Coroutine[None, None, T]]
    | Callable[
        [Callable[P, Coroutine[None, None, T]]],
        Callable[P, Coroutine[None, None, T]],
    ]
):
    """Retry if a transaction failed.

    If the wrapped method fails with an `sqlalchemy.exc.DBAPIError` exception,
    it is retried up to ``max_tries`` times. Any method with this decorator
    must be idempotent, since it may be re-run multiple times.

    One common use for this decorator is when the database engine has been
    configured with the ``REPEATABLE READ`` transaction isolation level and
    multiple writers may be updating the same object at the same time. The
    loser of the transaction race will raise one of the above exceptions, and
    can be retried with this decorator.

    Parameters
    ----------
    delay
        How long, in seconds, to wait between tries.
    max_tries
        Number of times to retry the transaction. Practical experience with
        ``REPEATABLE READ`` isolation suggests a count of at least three.

    Examples
    --------
    This decorator can be used with any idempotent Python function or method
    that makes database calls and should be retried on the above-listed
    exceptions.

    .. code-block:: python

       from safir.database import retry_async_transaction
       from sqlalchemy.ext.asyncio import async_scoped_session


       @retry_async_transaction(max_tries=5)
       def make_some_database_call(session: async_scoped_session) -> None: ...

    If the default value of ``max_tries`` is acceptable, this decorator can
    be used without arguments.

    .. code-block:: python

       from safir.database import retry_async_transaction
       from sqlalchemy.ext.asyncio import async_scoped_session


       @retry_async_transaction
       def make_some_database_call(session: async_scoped_session) -> None: ...
    """

    def decorator(
        f: Callable[P, Coroutine[None, None, T]],
    ) -> Callable[P, Coroutine[None, None, T]]:
        @wraps(f)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            for _ in range(1, max_tries):
                with contextlib.suppress(DBAPIError):
                    return await f(*args, **kwargs)
                await asyncio.sleep(delay)
            return await f(*args, **kwargs)

        return wrapper

    if __func is not None:
        return decorator(__func)
    else:
        return decorator
